# ----------------------------------------------------------------------
# |
# |  SubprocessEx.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-22 11:07:58
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Enhancements for the subprocess library."""

import copy
import os
import subprocess
import textwrap

from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, cast, IO, Optional

from dbrownell_Common.ContextlibEx import ExitStack
from dbrownell_Common.Streams.Capabilities import Capabilities
from dbrownell_Common.Streams.StreamDecorator import TextWriterT


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
@dataclass
class RunResult:
    """Result of running a process."""

    # ----------------------------------------------------------------------
    returncode: int
    output: str

    error_command_line: Optional[str] = field(default=None)

    # ----------------------------------------------------------------------
    def __post_init__(self):
        assert self.error_command_line is None or self.returncode != 0

    # ----------------------------------------------------------------------
    def RaiseOnError(self) -> None:
        if self.returncode != 0:
            assert self.error_command_line is not None

            raise Exception(
                textwrap.dedent(
                    """\
                    Command Line
                    ------------
                    {}

                    Output
                    ------
                    {}
                    """,
                ).format(
                    self.error_command_line.rstrip(),
                    self.output.rstrip(),
                ),
            )


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def Run(
    command_line: str,
    cwd: Optional[Path] = None,
    env: Optional[dict[str, str]] = None,
    *,
    supports_colors: Optional[bool] = None,
) -> RunResult:
    """Runs a command line and returns the result."""

    env_args: dict[str, str] = {
        Capabilities.SIMULATE_TERMINAL_INTERACTIVE_ENV_VAR: "0",
        Capabilities.SIMULATE_TERMINAL_HEADLESS_ENV_VAR: "1",
    }

    if supports_colors is not None:
        env_args[Capabilities.SIMULATE_TERMINAL_COLORS_ENV_VAR] = "1" if supports_colors else "0"

    result = subprocess.run(
        command_line,
        check=False,
        cwd=cwd,
        env=_SetEnvironment(env, **env_args),
        shell=True,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
    )

    content = result.stdout.decode("utf-8")
    content = content.replace("\r\n", "\n")

    return RunResult(
        result.returncode,
        content,
        command_line if result.returncode != 0 else None,
    )


# ----------------------------------------------------------------------
# pylint: disable=too-many-locals
def Stream(
    command_line: str,
    stream: TextWriterT,
    cwd: Optional[Path] = None,
    env: Optional[dict[str, str]] = None,
    *,
    stdin: Optional[str] = None,
    line_delimited_output: bool = False,  # Set to True to buffer lines
    is_headless: Optional[bool] = None,
    is_interactive: Optional[bool] = None,
    supports_colors: Optional[bool] = None,
) -> int:
    output_func = cast(Callable[[str], None], stream.write)
    flush_func = stream.flush

    capabilities = Capabilities.Get(stream)

    # Windows seems to want to interpret '\r\n' as '\n\n' when output is redirected to a file. Work
    # around that issue as best as we can.
    convert_newlines = False

    if capabilities.is_interactive:
        # Windows must always convert newlines
        convert_newlines = os.name.lower() == "nt"

    if convert_newlines:
        newline_original_output_func = output_func

        # ----------------------------------------------------------------------
        def NewlineOutput(
            content: str,
        ) -> None:
            newline_original_output_func(content.replace("\r\n", "\n"))

        # ----------------------------------------------------------------------

        output_func = NewlineOutput

    if line_delimited_output:
        line_delimited_original_output_func = output_func
        line_delimited_original_flush_func = flush_func

        cached_content: list[str] = []

        # ----------------------------------------------------------------------
        def LineDelimitedOutput(
            content: str,
        ) -> None:
            if content.endswith("\n"):
                content = "{}{}".format("".join(cached_content), content)
                cached_content[:] = []

                line_delimited_original_output_func(content)
            else:
                cached_content.append(content)

        # ----------------------------------------------------------------------
        def LineDelimitedFlush() -> None:
            if cached_content:
                content = "".join(cached_content)
                cached_content[:] = []
            else:
                content = ""

            if not content.endswith("\n"):
                content += "\n"

            line_delimited_original_output_func(content)
            line_delimited_original_flush_func()

        # ----------------------------------------------------------------------

        output_func = LineDelimitedOutput
        flush_func = LineDelimitedFlush

    if is_headless is None:
        is_headless = capabilities.is_headless
    if is_interactive is None:
        is_interactive = capabilities.is_interactive
    if supports_colors is None:
        supports_colors = capabilities.supports_colors

    with subprocess.Popen(
        command_line,
        cwd=cwd,
        env=_SetEnvironment(
            env,
            **{
                "PYTHONUNBUFFERED": "1",
                "COLUMNS": capabilities.columns,
                Capabilities.SIMULATE_TERMINAL_HEADLESS_ENV_VAR: "1" if is_headless else "0",
                Capabilities.SIMULATE_TERMINAL_INTERACTIVE_ENV_VAR: "1" if is_interactive else "0",
                Capabilities.SIMULATE_TERMINAL_COLORS_ENV_VAR: "1" if supports_colors else "0",
            },
        ),
        shell=True,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    ) as result:
        try:
            with ExitStack(flush_func):
                if stdin is not None:
                    assert result.stdin is not None

                    result.stdin.write(stdin.encode("utf-8"))
                    result.stdin.flush()
                    result.stdin.close()

                assert result.stdout is not None

                _ReadStateMachine.Execute(
                    result.stdout,
                    output_func,
                    convert_newlines=convert_newlines,
                )

                result_code = result.wait() or 0

        except IOError:  # pragma: no cover
            result_code = -1  # pragma: no cover

        return result_code


# ----------------------------------------------------------------------
# |
# |  Private Types
# |
# ----------------------------------------------------------------------
class _ReadStateMachine:
    """Reads content produced by a stream, ensuring and ansi escape sequences are properly grouped."""

    # ----------------------------------------------------------------------
    @classmethod
    def Execute(
        cls,
        input_stream: IO[bytes],
        output_func: Callable[[str], None],
        *,
        convert_newlines: bool,
    ) -> None:
        machine = cls(
            input_stream,
            convert_newlines=convert_newlines,
        )

        while True:
            if machine._buffered_input is not None:
                input_data = machine._buffered_input
                machine._buffered_input = None
            else:
                stream_data = machine._input_stream.read(1)
                if not stream_data:
                    break

                assert isinstance(stream_data, (str, bytes)), stream_data
                input_data = ord(stream_data)

            result = machine._process_func(input_data)
            if result is None:
                continue

            output_func(machine._ToString(result))

        if machine._buffered_output:
            output_func(machine._ToString(machine._buffered_output))

    # ----------------------------------------------------------------------
    def __init__(
        self,
        input_stream: IO[bytes],
        *,
        convert_newlines: bool,
    ):
        self._input_stream = input_stream
        self._convert_newlines = convert_newlines

        self._process_func: Callable[[int], Optional[list[int]]] = self._ProcessStandard

        self._buffered_input: Optional[int] = None
        self._buffered_output: list[int] = []

    # ----------------------------------------------------------------------
    # |
    # |  Private Types
    # |
    # ----------------------------------------------------------------------
    _a = ord("a")
    _z = ord("z")
    _A = ord("A")
    _Z = ord("Z")

    # ----------------------------------------------------------------------
    # |
    # |  Private Methods
    # |
    # ----------------------------------------------------------------------
    @classmethod
    def _IsAsciiLetter(
        cls,
        value: int,
    ) -> bool:
        return (cls._a <= value <= cls._z) or (cls._A <= value <= cls._Z)  # pragma: no cover

    # ----------------------------------------------------------------------
    def _IsNewlineish(
        self,
        value: int,
    ) -> bool:
        return self._convert_newlines and value in [
            10,  # '\r'
            13,  # '\n'
        ]

    # ----------------------------------------------------------------------
    @staticmethod
    def _IsEscape(
        value: int,
    ) -> bool:
        return value == 27

    # ----------------------------------------------------------------------
    @staticmethod
    def _ToString(
        value: list[int],
    ) -> str:
        if len(value) == 1:
            return chr(value[0])

        result = bytearray(value)

        # Attempt to decode as utf-8
        try:
            return result.decode("utf-8")
        except (UnicodeDecodeError, LookupError):  # pragma: no cover
            pass  # pragma: no cover

        raise Exception("The content '{}' could not be decoded.".format(result))  # pragma: no cover

    # ----------------------------------------------------------------------
    def _ProcessStandard(
        self,
        value: int,
    ) -> Optional[list[int]]:
        assert not self._buffered_output

        if self.__class__._IsEscape(value):  # pylint: disable=protected-access
            self._process_func = self._ProcessEscape
            self._buffered_output.append(value)

            return None

        if self._IsNewlineish(value):
            self._process_func = self._ProcessLineReset
            self._buffered_output.append(value)

            return None

        if value >> 6 == 0b11:
            # This is the first char of a multi-byte sequence
            self._process_func = self._ProcessMultiByte
            self._buffered_output.append(value)

            return None

        return [
            value,
        ]

    # ----------------------------------------------------------------------
    def _ProcessEscape(
        self,
        value: int,
    ) -> Optional[list[int]]:
        assert self._buffered_output
        self._buffered_output.append(value)

        if not self.__class__._IsAsciiLetter(value):  # pylint: disable=protected-access
            return None

        self._process_func = self._ProcessStandard

        return self._FlushBufferedOutput()

    # ----------------------------------------------------------------------
    def _ProcessLineReset(
        self,
        value: int,
    ) -> Optional[list[int]]:
        assert self._buffered_output

        if self._IsNewlineish(value):
            self._buffered_output.append(value)
            return None

        self._process_func = self._ProcessStandard

        assert self._buffered_input is None
        self._buffered_input = value

        return self._FlushBufferedOutput()

    # ----------------------------------------------------------------------
    def _ProcessMultiByte(
        self,
        value: int,
    ) -> Optional[list[int]]:
        assert self._buffered_output

        if value >> 6 == 0b10:
            # Continuation char
            self._buffered_output.append(value)
            return None

        self._process_func = self._ProcessStandard

        assert self._buffered_input is None
        self._buffered_input = value

        return self._FlushBufferedOutput()

    # ----------------------------------------------------------------------
    def _FlushBufferedOutput(self) -> Optional[list[int]]:
        assert self._buffered_output

        content = self._buffered_output
        self._buffered_output = []

        return content


# ----------------------------------------------------------------------
# |
# |  Private Functions
# |
# ----------------------------------------------------------------------
def _SetEnvironment(
    env: Optional[dict[str, str]],
    **kwargs: Any,
) -> dict[str, str]:
    if env is None:
        env = copy.deepcopy(os.environ)  # type: ignore

    assert env is not None

    for k, v in kwargs.items():
        env[k] = str(v) if not isinstance(v, str) else v

    if "PYTHONIOENCODING" not in env:
        env["PYTHONIOENCODING"] = "utf-8"

    if "COLUMNS" not in env:
        env["COLUMNS"] = str(Capabilities.DEFAULT_COLUMNS)

    return env



# ----------------------------------------------------------------------
# |
# |  ContextlibEx.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-19 07:08:25
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Enhancements to the contextlib package"""

from contextlib import contextmanager, ExitStack as ExitStackImpl
from typing import Any, Callable, Iterator


# ----------------------------------------------------------------------
@contextmanager
def ExitStack(
    *args: Callable[[], Any],
) -> Iterator[Any]:
    """Utility that invokes all of the provided functions when it goes out of scope."""

    with ExitStackImpl() as exit_stack:
        for arg in args:
            exit_stack.callback(arg)

        yield exit_stack



# ----------------------------------------------------------------------
# |
# |  ExecuteTasks.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-12 20:04:41
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Contains functionality to execute multiple tasks in parallel."""

import datetime
import multiprocessing
import shutil
import sys
import threading
import time
import traceback

from abc import abstractmethod, ABC
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, cast, Iterator, Optional, Protocol, TypeVar, Union
from unittest.mock import MagicMock

from rich.progress import Progress, TaskID, TimeElapsedColumn

from dbrownell_Common.ContextlibEx import ExitStack
from dbrownell_Common.InflectEx import inflect
from dbrownell_Common import PathEx
from dbrownell_Common.Streams.Capabilities import Capabilities
from dbrownell_Common.Streams.DoneManager import DoneManager
from dbrownell_Common import TextwrapEx


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
CATASTROPHIC_TASK_FAILURE_RESULT = -123

DISPLAY_COLUMN_WIDTH = int(Capabilities.DEFAULT_COLUMNS * 0.50)
STATUS_COLUMN_WIDTH = int(Capabilities.DEFAULT_COLUMNS * 0.30)


# ----------------------------------------------------------------------
class TransformException(Exception):
    """Exception raised when the Transform function encounters errors when processing a task."""

    pass  # pylint: disable=unnecessary-pass


# ----------------------------------------------------------------------
@dataclass
class TaskData:
    """Data associated with a single task."""

    # ----------------------------------------------------------------------
    display: str
    context: Any

    # Set this value if the task needs to be processed exclusively with respect to
    # other `TaskData` objects with the same execution lock.
    execution_lock: Optional[threading.Lock] = field(default=None)

    # The following values will be populated during task execution
    result: int = field(init=False)
    short_desc: Optional[str] = field(init=False)

    execution_time: datetime.timedelta = field(init=False)
    log_filename: Path = field(init=False)


# ----------------------------------------------------------------------
class Status(ABC):
    """Interface to set information about a single task."""

    # ----------------------------------------------------------------------
    @abstractmethod
    def SetTitle(
        self,
        title: str,
    ) -> None:
        raise Exception("Abstract method")  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnProgress(
        self,
        zero_based_step: Optional[int],
        status: Optional[str],
    ) -> bool:
        raise Exception("Abstract method")  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def OnInfo(
        self,
        value: str,
        *,
        verbose: bool = False,
    ) -> None:
        raise Exception("Abstract method")  # pragma: no cover


# ----------------------------------------------------------------------
class ExecuteTasksTypes:
    """Types used by ExecuteTasks."""

    # ----------------------------------------------------------------------
    class InitFuncType(Protocol):
        """Initializes a task for execution."""

        def __call__(
            self,
            context: Any,  # TaskData.context
        ) -> tuple[
            Path,
            "ExecuteTasksTypes.PrepareFuncType",
        ]:  # (Log filename, PrepareFuncType)
            ...  # pragma: no cover

    # ----------------------------------------------------------------------
    class PrepareFuncType(Protocol):
        """Prepares a task for execution."""

        def __call__(
            self,
            on_simple_status_func: Callable[
                [
                    str,  # Status
                ],
                None,
            ],
        ) -> Union[
            tuple[
                int,  # Number of steps to execute
                "ExecuteTasksTypes.ExecuteFuncType",
            ],
            "ExecuteTasksTypes.ExecuteFuncType",
        ]: ...  # pragma: no cover

    # ----------------------------------------------------------------------
    class ExecuteFuncType(Protocol):
        """Executes a task."""

        def __call__(
            self,
            status: Status,
        ) -> Union[
            tuple[
                int,  # Return code
                Optional[str],  # Final status message
            ],
            int,  # Return code
        ]: ...  # pragma: no cover


# ----------------------------------------------------------------------
@dataclass
class TransformResultComplete:
    """Complex result returned by a transform function."""

    value: Any
    return_code: Optional[int] = field(default=None)
    short_desc: Optional[str] = field(default=None)


# ----------------------------------------------------------------------
TransformedType = TypeVar(  # pylint: disable=invalid-name, typevar-name-incorrect-variance
    "TransformedType", covariant=True
)


# ----------------------------------------------------------------------
class TransformTasksExTypes:
    """Types used by TransformTasksEx."""

    # ----------------------------------------------------------------------
    class PrepareFuncType(Protocol[TransformedType]):
        """Prepares a task for transformation."""

        def __call__(
            self,
            context: Any,  # TaskData.context
            on_simple_status_func: Callable[
                [
                    str,  # Status
                ],
                None,
            ],
        ) -> Union[
            tuple[
                int,  # Number of steps to execute
                "TransformTasksExTypes.TransformFuncType",
            ],
            "TransformTasksExTypes.TransformFuncType",
        ]: ...  # pragma: no cover

    # ----------------------------------------------------------------------
    class TransformFuncType(Protocol[TransformedType]):
        """Transforms a task."""

        def __call__(
            self,
            status: Status,
        ) -> TransformResultComplete | TransformedType: ...  # pragma: no cover


# ----------------------------------------------------------------------
class TransformTasksTypes:
    """Types used by TransformTasks."""

    # ----------------------------------------------------------------------
    class TransformFuncType(Protocol[TransformedType]):
        """Transforms a task."""

        def __call__(
            self,
            context: Any,
            status: Status,
        ) -> TransformResultComplete | TransformedType: ...  # pragma: no cover


# ----------------------------------------------------------------------
class YieldQueueExecutorTypes:
    """Types used by YieldQueueExecutor."""

    # ----------------------------------------------------------------------
    class PrepareFuncType(Protocol):
        """Prepares a task for execution."""

        def __call__(
            self,
            on_simple_status_func: Callable[
                [
                    str,  # Status
                ],
                None,
            ],
        ) -> Union[
            tuple[
                int,  # Number of steps to execute
                "YieldQueueExecutorTypes.ExecuteFuncType",
            ],
            "YieldQueueExecutorTypes.ExecuteFuncType",
        ]: ...  # pragma: no cover

    # ----------------------------------------------------------------------
    class ExecuteFuncType(Protocol):
        """Executes a task."""

        def __call__(
            self,
            status: Status,
        ) -> Optional[str]:  # Status message
            ...  # pragma: no cover

    # ----------------------------------------------------------------------
    class EnqueueFuncType(Protocol):
        """Enqueues a task to execute."""

        def __call__(
            self,
            description: str,
            prepare_func: "YieldQueueExecutorTypes.PrepareFuncType",
        ) -> None: ...  # pragma: no cover


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def ExecuteTasks(
    dm: DoneManager,
    desc: str,
    tasks: list[TaskData],
    init_func: ExecuteTasksTypes.InitFuncType,
    *,
    quiet: bool = False,
    max_num_threads: Optional[int] = None,
    refresh_per_second: Optional[float] = None,
) -> None:
    """Executes tasks that each output to individual log files."""

    with _GenerateStatusInfo(
        len(tasks),
        dm,
        desc,
        tasks,
        quiet=quiet,
        refresh_per_second=refresh_per_second,
    ) as (status_factories, on_task_complete_func):
        if max_num_threads == 1 or len(tasks) == 1:
            for task_data, status_factory in zip(tasks, status_factories):
                with ExitStack(status_factory.Stop):
                    _ExecuteTask(
                        desc,
                        task_data,
                        init_func,
                        status_factory,
                        on_task_complete_func,
                        is_debug=dm.is_debug,
                    )

            return

        with ThreadPoolExecutor(
            max_workers=max_num_threads,
        ) as executor:
            # ----------------------------------------------------------------------
            def Impl(
                task_data: TaskData,
                status_factory: "_StatusFactory",
            ) -> None:
                with ExitStack(status_factory.Stop):
                    _ExecuteTask(
                        desc,
                        task_data,
                        init_func,
                        status_factory,
                        on_task_complete_func,
                        is_debug=dm.is_debug,
                    )

            # ----------------------------------------------------------------------

            futures = [
                executor.submit(Impl, task_data, status_factory)
                for task_data, status_factory in zip(tasks, status_factories)
            ]

            for future in futures:
                future.result()


# ----------------------------------------------------------------------
def TransformTasksEx(
    dm: DoneManager,
    desc: str,
    tasks: list[TaskData],
    prepare_func: TransformTasksExTypes.PrepareFuncType[TransformedType],
    *,
    quiet: bool = False,
    max_num_threads: Optional[int] = None,
    refresh_per_second: Optional[float] = None,
    no_compress_tasks: bool = False,
    return_exceptions: bool = False,
) -> list[
    Union[
        None,
        TransformedType,
        Exception,  # If `return_exceptions` is True and an exception was encountered
    ],
]:
    """Executes functions that return values; use this variation for tasks that are transformed in multiple steps or require custom preparation."""

    with _YieldTemporaryDirectory(dm) as temp_directory:
        cpu_count = multiprocessing.cpu_count()

        num_threads = min(len(tasks), cpu_count)
        if max_num_threads:
            num_threads = min(num_threads, max_num_threads)

        if no_compress_tasks or num_threads < cpu_count or num_threads == 1:
            impl_func = _TransformNotCompressed
        else:
            impl_func = _TransformCompressed

        return impl_func(
            temp_directory,
            dm,
            desc,
            tasks,
            prepare_func,
            quiet=quiet,
            num_threads=num_threads,
            refresh_per_second=refresh_per_second,
            return_exceptions=return_exceptions,
        )


# ----------------------------------------------------------------------
def TransformTasks(
    dm: DoneManager,
    desc: str,
    tasks: list[TaskData],
    transform_func: TransformTasksTypes.TransformFuncType[TransformedType],
    *,
    quiet: bool = False,
    max_num_threads: Optional[int] = None,
    refresh_per_second: Optional[float] = None,
    no_compress_tasks: bool = False,
    return_exceptions: bool = False,
) -> list[
    Union[
        None,
        TransformedType,
        Exception,  # If `return_exceptions` is True and an exception was encountered
    ],
]:
    """Executes functions that return values; use this function for single-step tasks that do not require extensive preparation."""

    # ----------------------------------------------------------------------
    def Prepare(
        context: Any,
        on_simple_status_func: Callable[[str], None],  # pylint: disable=unused-argument
    ) -> TransformTasksExTypes.TransformFuncType:
        return lambda status: transform_func(context, status)

    # ----------------------------------------------------------------------

    return TransformTasksEx(
        dm,
        desc,
        tasks,
        Prepare,
        quiet=quiet,
        max_num_threads=max_num_threads,
        refresh_per_second=refresh_per_second,
        no_compress_tasks=no_compress_tasks,
        return_exceptions=return_exceptions,
    )


# ----------------------------------------------------------------------
@contextmanager
def YieldQueueExecutor(
    dm: DoneManager,
    desc: str,
    *,
    quiet: bool = False,
    max_num_threads: Optional[int] = None,
    refresh_per_second: Optional[float] = None,
) -> Iterator[YieldQueueExecutorTypes.EnqueueFuncType]:
    """Yields a callable that can be used to enqueue tasks executed by workers running across multiple threads."""

    with _YieldTemporaryDirectory(dm) as temp_directory:
        num_threads = max_num_threads or multiprocessing.cpu_count()

        with _GenerateStatusInfo(
            None,
            dm,
            desc,
            [TaskData("", thread_index) for thread_index in range(num_threads)],
            quiet=quiet,
            refresh_per_second=refresh_per_second,
        ) as (status_factories, on_task_complete_func):
            queue: list[tuple[str, YieldQueueExecutorTypes.PrepareFuncType]] = []
            queue_lock = threading.Lock()

            queue_semaphore = threading.Semaphore(0)
            quit_event = threading.Event()

            # ----------------------------------------------------------------------
            def Enqueue(
                description: str,
                prepare_func: YieldQueueExecutorTypes.PrepareFuncType,
            ) -> None:
                with queue_lock:
                    queue.append((description, prepare_func))
                    queue_semaphore.release()

            # ----------------------------------------------------------------------
            def Impl(
                thread_index: int,
            ) -> None:
                log_filename = temp_directory / "{:06}.log".format(thread_index)
                status_factory = status_factories[thread_index]

                with ExitStack(status_factory.Stop):
                    while True:
                        with queue_semaphore:
                            with queue_lock:
                                if not queue:
                                    assert quit_event.is_set()
                                    break

                                task_desc, prepare_func = queue.pop(0)

                            # ----------------------------------------------------------------------
                            def Init(  # pylint: disable=unused-argument
                                *args,
                                **kwargs,
                            ) -> tuple[Path, ExecuteTasksTypes.PrepareFuncType]:  # pylint: disable=unused-argument
                                # ----------------------------------------------------------------------
                                def Prepare(
                                    on_simple_status_func: Callable[[str], None],
                                ) -> Union[
                                    tuple[int, ExecuteTasksTypes.ExecuteFuncType],
                                    ExecuteTasksTypes.ExecuteFuncType,
                                ]:
                                    prepare_result = prepare_func(on_simple_status_func)

                                    num_steps: Optional[int] = None
                                    execute_func: Optional[YieldQueueExecutorTypes.ExecuteFuncType] = None

                                    if isinstance(prepare_result, tuple):
                                        num_steps, execute_func = prepare_result
                                    else:
                                        execute_func = prepare_result

                                    assert execute_func is not None

                                    # ----------------------------------------------------------------------
                                    def Execute(
                                        status: Status,
                                    ) -> tuple[int, Optional[str]]:
                                        return 0, execute_func(status)

                                    # ----------------------------------------------------------------------

                                    if num_steps is not None:
                                        return num_steps, Execute

                                    return Execute

                                # ----------------------------------------------------------------------

                                return log_filename, Prepare

                            # ----------------------------------------------------------------------

                            _ExecuteTask(
                                desc,
                                TaskData(task_desc, None),
                                Init,
                                status_factory,
                                on_task_complete_func,
                                is_debug=dm.is_debug,
                            )

            # ----------------------------------------------------------------------

            with ThreadPoolExecutor(
                max_workers=num_threads,
            ) as executor:
                futures = [executor.submit(Impl, thread_index) for thread_index in range(num_threads)]

                yield Enqueue

                quit_event.set()
                queue_semaphore.release(num_threads)

                for future in futures:
                    future.result()


# ----------------------------------------------------------------------
# |
# |  Private Types
# |
# ----------------------------------------------------------------------
class _InternalStatus(Status):
    # ----------------------------------------------------------------------
    @abstractmethod
    def SetNumSteps(
        self,
        num_steps: int,
    ) -> None:
        raise Exception("Abstract method")  # pragma: no cover


# ----------------------------------------------------------------------
class _StatusFactory(ABC):
    """Interface for object that is able to create Status objects."""

    # ----------------------------------------------------------------------
    @abstractmethod
    @contextmanager
    def CreateStatus(
        self,
        display: str,
    ) -> Iterator[_InternalStatus]:
        raise Exception("Abstract method")  # pragma: no cover

    # ----------------------------------------------------------------------
    @abstractmethod
    def Stop(self) -> None:
        raise Exception("Abstract method")  # pragma: no cover


# ----------------------------------------------------------------------
# |
# |  Private Functions
# |
# ----------------------------------------------------------------------
@contextmanager
def _GenerateStatusInfo(
    num_tasks_display_value: Optional[int],
    dm: DoneManager,
    desc: str,
    tasks: list[TaskData],
    *,
    quiet: bool,
    refresh_per_second: Optional[float],
) -> Iterator[
    tuple[
        list[_StatusFactory],
        Callable[[TaskData], None],
    ],
]:
    success_count = 0
    error_count = 0
    warning_count = 0

    count_lock = threading.Lock()

    # Create the heading
    if desc.endswith("..."):
        desc = desc[: -len("...")]

    heading = desc

    if num_tasks_display_value is not None:
        items_text = inflect.no("item", num_tasks_display_value)

        if heading:
            heading += " ({})".format(items_text)
        else:
            heading = items_text

    heading += "..."

    with dm.Nested(
        heading,
        [
            lambda: "{} succeeded".format(inflect.no("item", success_count)),
            lambda: "{} with errors".format(inflect.no("item", error_count)),
            lambda: "{} with warnings".format(inflect.no("item", warning_count)),
        ],
    ) as execute_dm:
        # ----------------------------------------------------------------------
        def OnTaskDataComplete(
            task_data: TaskData,
        ) -> tuple[int, int, int]:
            nonlocal success_count
            nonlocal error_count
            nonlocal warning_count

            with count_lock:
                if task_data.result < 0:
                    error_count += 1

                    if execute_dm.result >= 0:
                        execute_dm.result = task_data.result

                elif task_data.result > 0:
                    warning_count += 1

                    if execute_dm.result == 0:
                        execute_dm.result = task_data.result

                else:
                    success_count += 1

            return success_count, error_count, warning_count

        # ----------------------------------------------------------------------

        with (_GenerateProgressStatusInfo if dm.capabilities.is_interactive else _GenerateNoopStatusInfo)(
            num_tasks_display_value,
            execute_dm,
            tasks,
            OnTaskDataComplete,
            quiet=quiet,
            refresh_per_second=refresh_per_second,
        ) as value:
            yield value


# ----------------------------------------------------------------------
@contextmanager
def _GenerateProgressStatusInfo(
    num_tasks_display_value: Optional[int],
    dm: DoneManager,
    tasks: list[TaskData],
    on_task_complete_func: Callable[[TaskData], tuple[int, int, int]],
    *,
    quiet: bool,
    refresh_per_second: Optional[float],
) -> Iterator[
    tuple[
        list[_StatusFactory],
        Callable[[TaskData], None],
    ],
]:
    with dm.YieldStdout() as stdout_context:
        stdout_context.persist_content = False

        # Technically speaking, it would be more correct to use `stdout_context.stream` here
        # rather than referencing `sys.stdout` directly, but it is really hard to work with mocked
        # stream as mocks will create mocks for everything called on the mock. Use sys.stdout
        # directly to avoid that particular problem.
        assert stdout_context.stream is sys.stdout or isinstance(stdout_context.stream, MagicMock), (
            stdout_context.stream
        )

        progress_bar = Progress(
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            "{task.fields[status]}",
            console=Capabilities.Get(sys.stdout).CreateRichConsole(sys.stdout),
            transient=True,
            refresh_per_second=refresh_per_second or 10,
        )

        # ----------------------------------------------------------------------
        def CreateDescription(
            value: str,
            *,
            indent: bool = True,
        ) -> str:
            return TextwrapEx.BoundedLJust(
                "{}{}{}".format(
                    stdout_context.line_prefix,
                    "  " if indent else "",
                    value,
                ),
                DISPLAY_COLUMN_WIDTH,
            )

        # ----------------------------------------------------------------------
        class StatusImpl(_InternalStatus):  # pylint: disable=missing-class-docstring
            # ----------------------------------------------------------------------
            def __init__(
                self,
                task_id: TaskID,
            ):
                self._task_id = task_id

                self._num_steps: Optional[int] = None
                self._current_step: Optional[int] = None

            # ----------------------------------------------------------------------
            def SetNumSteps(
                self,
                num_steps: int,
            ) -> None:
                assert self._num_steps is None
                assert self._current_step is None

                self._num_steps = num_steps
                self._current_step = 0

                progress_bar.update(
                    self._task_id,
                    completed=self._current_step,
                    refresh=False,
                    total=self._num_steps,
                )

            # ----------------------------------------------------------------------
            def SetTitle(
                self,
                title: str,
            ) -> None:
                progress_bar.update(
                    self._task_id,
                    description=CreateDescription(title),
                    refresh=False,
                )

            # ----------------------------------------------------------------------
            def OnProgress(
                self,
                zero_based_step: Optional[int],
                status: Optional[str],
            ) -> bool:
                if zero_based_step is not None:
                    assert self._num_steps is not None
                    self._current_step = zero_based_step

                status = status or ""

                if self._num_steps is not None:
                    assert self._current_step is not None

                    status = "({} of {}) {}".format(
                        self._current_step + 1,
                        self._num_steps,
                        status,
                    )

                status = TextwrapEx.BoundedLJust(status, STATUS_COLUMN_WIDTH)

                progress_bar.update(
                    self._task_id,
                    completed=self._current_step,
                    refresh=False,
                    status=status,
                )

                return True

            # ----------------------------------------------------------------------
            def OnInfo(
                self,
                value: str,
                *,
                verbose: bool = False,
            ) -> None:
                if verbose:
                    if not dm.is_verbose:
                        return

                    assert TextwrapEx.VERBOSE_COLOR_ON == "\033[;7m", "Ensure that the colors stay in sync"
                    prefix = "[black on white]VERBOSE:[/] "
                else:
                    assert TextwrapEx.INFO_COLOR_ON == "\033[;7m", "Ensure that the colors stay in sync"
                    prefix = "[black on white]INFO:[/] "

                progress_bar.print(
                    "{}{}{}".format(
                        stdout_context.line_prefix,
                        prefix,
                        value,
                    ),
                    highlight=False,
                )

                stdout_context.persist_content = True

        # ----------------------------------------------------------------------
        class StatusFactory(_StatusFactory):  # pylint: disable=missing-class-docstring
            # ----------------------------------------------------------------------
            def __init__(
                self,
                task_id: TaskID,
            ):
                self._task_id = task_id

            # ----------------------------------------------------------------------
            @contextmanager
            def CreateStatus(
                self,
                display: str,
            ) -> Iterator[_InternalStatus]:
                progress_bar.update(
                    self._task_id,
                    completed=0,
                    description=CreateDescription(display),
                    refresh=False,
                    status="",
                    total=None,
                    visible=not quiet,
                )

                progress_bar.start_task(self._task_id)
                with ExitStack(lambda: progress_bar.stop_task(self._task_id)):
                    yield StatusImpl(self._task_id)

            # ----------------------------------------------------------------------
            def Stop(self) -> None:
                progress_bar.update(
                    self._task_id,
                    refresh=False,
                    visible=False,
                )

        # ----------------------------------------------------------------------

        total_progress_id = progress_bar.add_task(
            CreateDescription(
                "" if num_tasks_display_value is None else "Total Progress",
                indent=False,
            ),
            status="",
            total=num_tasks_display_value,
            visible=num_tasks_display_value is not None,
        )

        # ----------------------------------------------------------------------
        def OnTaskComplete(
            task_data: TaskData,
        ) -> None:
            if not quiet and task_data.result != 0:
                if task_data.result < 0:
                    assert TextwrapEx.ERROR_COLOR_ON == "\033[31;1m", "Ensure that the colors stay in sync"
                    color = "red"
                    header = "ERROR"
                else:
                    assert TextwrapEx.WARNING_COLOR_ON == "\033[33;1m", "Ensure that the colors stay in sync"
                    color = "yellow"
                    header = "WARNING"

                if not task_data.log_filename.is_file():
                    suffix = ""
                else:
                    if dm.capabilities.is_headless:
                        suffix = str(task_data.log_filename)
                    else:
                        suffix = "[link=file:///{}]View Log[/]".format(task_data.log_filename.as_posix())

                    suffix = r" \[{}]".format(suffix)

                progress_bar.print(
                    r"{prefix}[bold {color}]{header}:[/] {name}: {result}{short_desc}{suffix}".format(
                        prefix=stdout_context.line_prefix,
                        color=color,
                        header=header,
                        name=task_data.display,
                        result=task_data.result,
                        short_desc=(" ({})".format(task_data.short_desc) if task_data.short_desc else ""),
                        suffix=suffix,
                    ),
                    highlight=False,
                )

                stdout_context.persist_content = True

            success_count, error_count, warning_count = on_task_complete_func(task_data)

            if Capabilities.Get(sys.stdout).supports_colors:
                success_on = TextwrapEx.SUCCESS_COLOR_ON
                error_on = TextwrapEx.ERROR_COLOR_ON
                warning_on = TextwrapEx.WARNING_COLOR_ON
                color_off = TextwrapEx.COLOR_OFF
            else:
                success_on = ""
                error_on = ""
                warning_on = ""
                color_off = ""

            parts: list[str] = []

            for color_on, count, suffix in [
                (success_on, success_count, "succeeded"),
                (error_on, error_count, "failed"),
                (warning_on, warning_count, inflect.plural_verb("warning", warning_count)),
            ]:
                if count == 0:
                    content = "0"
                else:
                    content = "{}{}{}".format(color_on, count, color_off)

                parts.append("{} {}".format(content, suffix))

            progress_bar.update(
                total_progress_id,
                advance=1,
                refresh=False,
                status=TextwrapEx.BoundedLJust(", ".join(parts), STATUS_COLUMN_WIDTH),
            )

        # ----------------------------------------------------------------------

        enqueueing_status = "{}Enqueueing tasks...".format(stdout_context.line_prefix)

        stdout_context.stream.write(enqueueing_status)
        stdout_context.stream.flush()

        status_factories: list[_StatusFactory] = []

        for task in tasks:
            status_factories.append(
                StatusFactory(
                    progress_bar.add_task(
                        CreateDescription(task.display),
                        start=False,
                        status="",
                        total=None,
                        visible=False,
                    ),
                ),
            )

        stdout_context.stream.write("\r{}\r".format(" " * len(enqueueing_status)))
        stdout_context.stream.flush()

        progress_bar.start()
        with ExitStack(progress_bar.stop):
            yield status_factories, OnTaskComplete


# ----------------------------------------------------------------------
@contextmanager
def _GenerateNoopStatusInfo(
    num_tasks_display_value: Optional[int],  # pylint: disable=unused-argument
    dm: DoneManager,
    tasks: list[TaskData],
    on_task_complete_func: Callable[[TaskData], tuple[int, int, int]],
    *,
    quiet: bool,
    refresh_per_second: Optional[float],  # pylint: disable=unused-argument
) -> Iterator[
    tuple[
        list[_StatusFactory],
        Callable[[TaskData], None],
    ],
]:
    # ----------------------------------------------------------------------
    class StatusImpl(_InternalStatus):  # pylint: disable=missing-class-docstring
        # ----------------------------------------------------------------------
        def SetNumSteps(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
            pass

        # ----------------------------------------------------------------------
        def SetTitle(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
            pass

        # ----------------------------------------------------------------------
        def OnProgress(self, *args, **kwargs) -> bool:  # pylint: disable=unused-argument
            return True

        # ----------------------------------------------------------------------
        def OnInfo(self, *args, **kwargs) -> None:  # pylint: disable=unused-argument
            pass

    # ----------------------------------------------------------------------
    class StatusFactory(_StatusFactory):  # pylint: disable=missing-class-docstring
        # ----------------------------------------------------------------------
        @contextmanager
        def CreateStatus(  # pylint: disable=unused-argument
            self,
            *args,
            **kwargs,
        ) -> Iterator[_InternalStatus]:
            yield StatusImpl()

        # ----------------------------------------------------------------------
        def Stop(self) -> None:
            pass

    # ----------------------------------------------------------------------
    def OnTaskComplete(
        task_data: TaskData,
    ) -> None:
        on_task_complete_func(task_data)

        if not quiet and task_data.result != 0 and task_data.log_filename.is_file():
            content = "{name}: {result}{short_desc} [{suffix}]\n".format(
                name=task_data.display,
                result=task_data.result,
                short_desc=" ({})".format(task_data.short_desc) if task_data.short_desc else "",
                suffix=(
                    str(task_data.log_filename)
                    if dm.capabilities.is_headless
                    else TextwrapEx.CreateAnsiHyperLink(
                        "file:///{}".format(task_data.log_filename.as_posix()),
                        "View Log",
                    )
                ),
            )

            if task_data.result < 0:
                dm.WriteError(content)
            else:
                dm.WriteWarning(content)

    # ----------------------------------------------------------------------

    yield (
        cast(list[_StatusFactory], [StatusFactory() for _ in tasks]),
        OnTaskComplete,
    )


# ----------------------------------------------------------------------
def _ExecuteTask(
    desc: str,
    task_data: TaskData,
    init_func: ExecuteTasksTypes.InitFuncType,
    status_factory: _StatusFactory,
    on_task_complete_func: Callable[[TaskData], None],
    *,
    is_debug: bool,
) -> None:
    with ExitStack(lambda: on_task_complete_func(task_data)):
        start_time = time.perf_counter()

        try:
            with status_factory.CreateStatus(task_data.display) as status:
                task_data.log_filename, prepare_func = init_func(task_data.context)

                # ----------------------------------------------------------------------
                def OnSimpleStatus(
                    value: str,
                ) -> None:
                    status.OnProgress(None, value)

                # ----------------------------------------------------------------------

                prepare_result = prepare_func(OnSimpleStatus)

                num_steps: Optional[int] = None
                execute_func: Optional[ExecuteTasksTypes.ExecuteFuncType] = None

                if isinstance(prepare_result, tuple):
                    num_steps, execute_func = prepare_result
                else:
                    execute_func = prepare_result

                assert execute_func is not None

                # ----------------------------------------------------------------------
                @contextmanager
                def AcquireExecutionLock():
                    if task_data.execution_lock is None:
                        yield
                        return

                    OnSimpleStatus("Waiting...")
                    with task_data.execution_lock:
                        yield

                # ----------------------------------------------------------------------

                with AcquireExecutionLock():
                    if num_steps is not None:
                        assert num_steps >= 0, num_steps
                        status.SetNumSteps(num_steps)

                    execute_result = execute_func(status)

                    if isinstance(execute_result, tuple):
                        task_data.result, task_data.short_desc = execute_result
                    else:
                        task_data.result = execute_result
                        task_data.short_desc = None

        except KeyboardInterrupt:  # pylint: disable=try-except-raise
            raise

        except Exception as ex:
            if is_debug:
                error = traceback.format_exc()
            else:
                error = str(ex)

            error = error.rstrip()

            if not hasattr(task_data, "log_filename"):
                # If here, this error has happened before we have received anything
                # from the initial callback. Create a log file and write the exception
                # information.
                task_data.log_filename = PathEx.CreateTempFileName()
                assert task_data.log_filename is not None

                with task_data.log_filename.open("w") as f:
                    f.write(error)

            else:
                with task_data.log_filename.open("a+") as f:
                    f.write("\n\n{}\n".format(error))

            if isinstance(ex, TransformException):
                result = 1
                short_desc = "{} failed".format(task_data.display)
            else:
                result = CATASTROPHIC_TASK_FAILURE_RESULT
                short_desc = "{} failed".format(desc)

            task_data.result = result
            task_data.short_desc = short_desc

        finally:
            assert hasattr(task_data, "result")
            assert hasattr(task_data, "short_desc")
            assert hasattr(task_data, "log_filename")

            task_data.execution_time = datetime.timedelta(seconds=time.perf_counter() - start_time)


# ----------------------------------------------------------------------
@contextmanager
def _YieldTemporaryDirectory(
    dm: DoneManager,
) -> Iterator[Path]:
    temp_directory = PathEx.CreateTempDirectory()

    # ----------------------------------------------------------------------
    def OnExit():
        # Remove the temp directory if everything worked as expected or there aren't any log files to view
        if dm.result == 0 or not any(temp_directory.iterdir()):
            shutil.rmtree(temp_directory)
            return

        if dm.capabilities.is_headless:
            dm.WriteInfo(
                "\nThe temporary working directory '{}' was preserved due to errors encountered while executing tasks.".format(
                    temp_directory
                )
            )
        else:
            dm.WriteInfo(
                "\nThe {} was preserved due to errors encountered while executing tasks.".format(
                    TextwrapEx.CreateAnsiHyperLink(
                        "file:///{}".format(temp_directory.as_posix()),
                        "temporary working directory",
                    ),
                ),
            )

    # ----------------------------------------------------------------------

    with ExitStack(OnExit):
        yield temp_directory


# ----------------------------------------------------------------------
def _TransformCompressed(
    temp_directory: Path,
    dm: DoneManager,
    desc: str,
    tasks: list[TaskData],
    prepare_func: TransformTasksExTypes.PrepareFuncType[TransformedType],
    *,
    quiet: bool,
    num_threads: int,
    refresh_per_second: Optional[float],
    return_exceptions: bool,
) -> list[
    Union[
        None,
        TransformedType,
        Exception,  # If `return_exceptions` is True and an exception was encountered
    ],
]:
    assert num_threads > 1, num_threads

    all_results: list[None | TransformedType | Exception] = [
        None,
    ] * len(tasks)

    with _GenerateStatusInfo(
        len(tasks),
        dm,
        desc,
        [TaskData("", thread_index) for thread_index in range(num_threads)],
        quiet=quiet,
        refresh_per_second=refresh_per_second,
    ) as (status_factories, on_task_complete_func):
        task_index = 0
        task_index_lock = threading.Lock()

        # ----------------------------------------------------------------------
        def Impl(
            thread_index: int,
        ) -> None:
            nonlocal task_index

            log_filename = temp_directory / "{:06}.log".format(thread_index)
            status_factory = status_factories[thread_index]

            with ExitStack(status_factory.Stop):
                while True:
                    with task_index_lock:
                        this_task_index = task_index
                        task_index += 1

                    if this_task_index >= len(tasks):
                        break

                    task_data = tasks[this_task_index]

                    # ----------------------------------------------------------------------
                    def Init(
                        *args,
                        **kwargs,  # pylint: disable=unused-argument
                    ) -> tuple[Path, ExecuteTasksTypes.PrepareFuncType]:
                        # ----------------------------------------------------------------------
                        def Prepare(
                            on_simple_status_func: Callable[[str], None],
                        ) -> Union[
                            tuple[int, ExecuteTasksTypes.ExecuteFuncType],
                            ExecuteTasksTypes.ExecuteFuncType,
                        ]:
                            prepare_result = prepare_func(task_data.context, on_simple_status_func)

                            num_steps: Optional[int] = None
                            transform_func: Optional[TransformTasksExTypes.TransformFuncType] = None

                            if isinstance(prepare_result, tuple):
                                num_steps, transform_func = prepare_result
                            else:
                                transform_func = prepare_result

                            assert transform_func is not None

                            # ----------------------------------------------------------------------
                            def Execute(
                                status: Status,
                            ) -> tuple[int, Optional[str]]:
                                try:
                                    transform_result = transform_func(status)

                                    return_code = 0
                                    result: Any = None
                                    short_desc: Optional[str] = None

                                    if isinstance(transform_result, TransformResultComplete):
                                        result = transform_result.value
                                        return_code = transform_result.return_code or 0
                                        short_desc = transform_result.short_desc
                                    else:
                                        result = transform_result

                                    all_results[this_task_index] = result
                                    return return_code, short_desc

                                except Exception as ex:
                                    if return_exceptions:
                                        all_results[this_task_index] = ex

                                    raise

                            # ----------------------------------------------------------------------

                            if num_steps is None:
                                return Execute

                            return num_steps, Execute

                        # ----------------------------------------------------------------------

                        return log_filename, Prepare

                    # ----------------------------------------------------------------------

                    _ExecuteTask(
                        desc,
                        task_data,
                        Init,
                        status_factory,
                        on_task_complete_func,
                        is_debug=dm.is_debug,
                    )

        # ----------------------------------------------------------------------

        with ThreadPoolExecutor(
            max_workers=num_threads,
        ) as executor:
            futures = [executor.submit(Impl, thread_index) for thread_index in range(num_threads)]

            for future in futures:
                future.result()

    return all_results


# ----------------------------------------------------------------------
def _TransformNotCompressed(
    temp_directory: Path,
    dm: DoneManager,
    desc: str,
    tasks: list[TaskData],
    prepare_func: TransformTasksExTypes.PrepareFuncType[TransformedType],
    *,
    quiet: bool,
    num_threads: int,
    refresh_per_second: Optional[float],
    return_exceptions: bool,
) -> list[
    Union[
        None,
        TransformedType,
        Exception,  # If `return_exceptions` is True and an exception was encountered
    ],
]:
    """Executes tasks with one executor per task."""

    # Update the task context with the task index
    for task_index, task in enumerate(tasks):
        task.context = (task_index, task.context)

    # ----------------------------------------------------------------------
    def RestoreTaskContext():
        for task in tasks:
            task.context = task.context[1]

    # ----------------------------------------------------------------------

    with ExitStack(RestoreTaskContext):
        all_results: list[None | TransformedType | Exception] = [
            None,
        ] * len(tasks)

        # ----------------------------------------------------------------------
        def Init(
            context: tuple[int, Any],
        ) -> tuple[Path, ExecuteTasksTypes.PrepareFuncType]:
            task_index, context = context

            log_filename = temp_directory / "{:06}.log".format(task_index)

            # ----------------------------------------------------------------------
            def Prepare(
                on_simple_status_func: Callable[[str], None],
            ) -> Union[
                tuple[int, ExecuteTasksTypes.ExecuteFuncType],
                ExecuteTasksTypes.ExecuteFuncType,
            ]:
                prepare_result = prepare_func(context, on_simple_status_func)

                num_steps: Optional[int] = None
                transform_func: Optional[TransformTasksExTypes.TransformFuncType] = None

                if isinstance(prepare_result, tuple):
                    num_steps, transform_func = prepare_result
                else:
                    transform_func = prepare_result

                assert transform_func is not None

                # ----------------------------------------------------------------------
                def Execute(
                    status: Status,
                ) -> tuple[int, Optional[str]]:
                    try:
                        transform_result = transform_func(status)

                        return_code = 0
                        result: Any = None
                        short_desc: Optional[str] = None

                        if isinstance(transform_result, TransformResultComplete):
                            result = transform_result.value
                            return_code = transform_result.return_code or 0
                            short_desc = transform_result.short_desc
                        else:
                            result = transform_result

                        all_results[task_index] = result
                        return return_code, short_desc

                    except Exception as ex:
                        if return_exceptions:
                            all_results[task_index] = ex

                        raise

                # ----------------------------------------------------------------------

                if num_steps is None:
                    return Execute

                return num_steps, Execute

            # ----------------------------------------------------------------------

            return log_filename, Prepare

        # ----------------------------------------------------------------------

        ExecuteTasks(
            dm,
            desc,
            tasks,
            Init,
            quiet=quiet,
            max_num_threads=num_threads,
            refresh_per_second=refresh_per_second,
        )

        return all_results



# ----------------------------------------------------------------------
# |
# |  DoneManager_UnitTest.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-06 19:17:30
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Unit tests for DoneManager.py."""

import re
import sys
import textwrap

from io import StringIO
from typing import cast, Optional

import pytest

from dbrownell_Common import TextwrapEx
from dbrownell_Common.Streams.Capabilities import Capabilities
from dbrownell_Common.Streams.DoneManager import DoneManager, DoneManagerException, Flags
from dbrownell_Common.Streams.TextWriter import TextWriter


# ----------------------------------------------------------------------
def test_Simple():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing...") as dm:
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_SimpleNoTrailingHeaderDots():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_SimpleTrailingHeaderNewline():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing...\n\n") as dm:
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_DoneSuffixNone():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", None):
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_DoneSuffixReturnString():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", lambda: "done_suffix"):
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>, done_suffix)\n"


# ----------------------------------------------------------------------
def test_DoneSuffixReturnNone():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", lambda: None):
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_DoneSuffixes():
    sink = _CreateSink()

    with DoneManager.Create(
        sink,
        "Testing",
        [
            lambda: "one",
            lambda: None,
            lambda: "three",
            lambda: "four",
        ],
    ):
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>, one, three, four)\n"


# ----------------------------------------------------------------------
def test_PrefixAndSuffixString():
    sink = _CreateSink()

    with DoneManager.Create(
        sink,
        "Testing",
        prefix="*\n__prefix__*",
        suffix="*\n__suffix__*",
    ) as dm:
        dm.WriteError("Line 1\nLine 2\n")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          ERROR: Line 1
                 Line 2
        *
        __prefix__*DONE! (-1, <Scrubbed Time>)
        *
        __suffix__*""",
    )


# ----------------------------------------------------------------------
def test_PrefixAndSuffixCallable():
    wrote_content = False

    # ----------------------------------------------------------------------
    def Prefix() -> str:
        return "*\n__prefix__*"

    # ----------------------------------------------------------------------
    def Suffix() -> str:
        return "*\n__suffix__*" if wrote_content else ""

    # ----------------------------------------------------------------------

    # wrote_content == False
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", prefix=Prefix, suffix=Suffix) as dm:
        pass

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...*
        __prefix__*DONE! (0, <Scrubbed Time>)
        """,
    )

    # wrote_content == True
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", prefix=Prefix, suffix=Suffix) as dm:
        dm.WriteLine("Line 1")
        wrote_content = True

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
        *
        __prefix__*DONE! (0, <Scrubbed Time>)
        *
        __suffix__*""",
    )


# ----------------------------------------------------------------------
def test_LinePrefixString():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", line_prefix="****") as dm:
        dm.WriteLine("Line 1\nLine 2")
        dm.WriteLine("Line 3")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
        ****Line 1
        ****Line 2
        ****Line 3
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_NoDisplay():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", display=False):
        pass

    assert _Scrub(sink.getvalue()) == "Testing"


# ----------------------------------------------------------------------
def test_NoDisplayResult():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", display_result=False):
        pass

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (<Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_NoDisplayTime():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", display_time=False):
        pass

    assert sink.getvalue() == "Testing...DONE! (0)\n"


# ----------------------------------------------------------------------
def test_WithErrorResult():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.result = -1

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (-1, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_WithWarningResult():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.result = 123

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (123, <Scrubbed Time>)\n"


# ----------------------------------------------------------------------
def test_DisplayExceptions():
    # display_exceptions == True
    sink = _CreateSink()

    with pytest.raises(
        Exception,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing"):
            raise Exception("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          ERROR: The exception
        DONE! (-1, <Scrubbed Time>)
        """,
    )

    # display_exceptions == False
    sink = _CreateSink()

    with pytest.raises(
        Exception,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing", display_exceptions=False):
            raise Exception("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_DisplayExceptionDetails():
    # display_exception_details == False
    sink = _CreateSink()

    with pytest.raises(
        Exception,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing", display_exception_details=False):
            raise Exception("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          ERROR: The exception
        DONE! (-1, <Scrubbed Time>)
        """,
    )

    # display_exception_details == True
    sink = _CreateSink()

    with pytest.raises(
        Exception,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing", display_exception_details=True):
            raise Exception("The exception")

    content = _Scrub(sink.getvalue())

    assert re.match(
        r"""(?#
        Header                  )Testing\.\.\.\n(?#
        Error                   )  ERROR: Traceback \(most recent call last\):\n(?#
        Traceback lines begin   )(?:(?#
          Traceback line        )         .+\n(?#
        Traceback lines end     ))+?(?#
        Last Traceback line     )         Exception: The exception\n+(?#
        Done                    )DONE! \(-1, <Scrubbed Time>\)\n(?#
        )""",
        content,
    ), content


# ----------------------------------------------------------------------
def test_DoneManagerException():
    sink = _CreateSink()

    with pytest.raises(
        DoneManagerException,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing", display_exception_details=True):
            raise DoneManagerException("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          ERROR: The exception
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_NestedException():
    sink = _CreateSink()

    with pytest.raises(
        Exception,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing") as dm:
            with dm.Nested("Nested"):
                raise Exception("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Nested...
            ERROR: The exception
          DONE! (-1, <Scrubbed Time>)
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_SuppressExceptions():
    # suppress_exceptions == False
    sink = _CreateSink()

    with pytest.raises(
        Exception,
        match=re.escape("The exception"),
    ):
        with DoneManager.Create(sink, "Testing", suppress_exceptions=False):
            raise Exception("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        textwrap.dedent(
            """\
            Testing...
              ERROR: The exception
            DONE! (-1, <Scrubbed Time>)
            """,
        ),
    )

    # suppress_exceptions == True
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", suppress_exceptions=True):
        raise Exception("The exception")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          ERROR: The exception
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_SuppressExceptionsNoDisplayExceptions():
    with pytest.raises(
        Exception,
        match=re.escape("Exceptions should not be suppressed when they are not displayed."),
    ):
        with DoneManager.Create(StringIO(), "Testing", display_exceptions=False, suppress_exceptions=True):
            pass


# ----------------------------------------------------------------------
def test_StandardFlags():
    with DoneManager.Create(_CreateSink(), "Testing", flags=Flags.Create(verbose=False, debug=False)) as dm:
        assert dm.is_verbose is False
        assert dm.is_debug is False


# ----------------------------------------------------------------------
def test_VerboseFlags():
    with DoneManager.Create(_CreateSink(), "Testing", flags=Flags.Create(verbose=True, debug=False)) as dm:
        assert dm.is_verbose is True
        assert dm.is_debug is False


# ----------------------------------------------------------------------
def test_DebugFlags():
    with DoneManager.Create(_CreateSink(), "Testing", flags=Flags.Create(verbose=False, debug=True)) as dm:
        assert dm.is_verbose is True
        assert dm.is_debug is True


# ----------------------------------------------------------------------
def test_Capabilities():
    sink = StringIO()
    capabilities = Capabilities(stream=sink)

    with DoneManager.Create(sink, "Testing") as dm:
        assert dm.capabilities is capabilities


# ----------------------------------------------------------------------
def test_isatty():
    with DoneManager.Create(_FakeStream(), "Testing") as dm:
        assert dm.isatty() is False

    with DoneManager.Create(_FakeStream(isatty=True), "Testing") as dm:
        assert dm.isatty() is True


# ----------------------------------------------------------------------
def test_WriteLine():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteLine("Line 1")
        dm.WriteLine("Line 2")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
          Line 2
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_WriteSuccess():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteSuccess("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          SUCCESS: Line 1
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_WriteSuccessWithColor():
    sink = _CreateSink(supports_colors=True)

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteSuccess("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          {color_on}SUCCESS:{color_off} Line 1
        DONE! ({color_on}0{color_off}, <Scrubbed Time>)
        """,
    ).format(
        color_on=TextwrapEx.SUCCESS_COLOR_ON,
        color_off=TextwrapEx.COLOR_OFF,
    )


# ----------------------------------------------------------------------
def test_WriteError():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteError("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          ERROR: Line 1
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_WriteErrorWithColor():
    sink = _CreateSink(supports_colors=True)

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteError("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          {color_on}ERROR:{color_off} Line 1
        DONE! ({color_on}-1{color_off}, <Scrubbed Time>)
        """,
    ).format(
        color_on=TextwrapEx.ERROR_COLOR_ON,
        color_off=TextwrapEx.COLOR_OFF,
    )


# ----------------------------------------------------------------------
def test_WriteWarning():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteWarning("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          WARNING: Line 1
        DONE! (1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_WriteWarningWithColor():
    sink = _CreateSink(supports_colors=True)

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteWarning("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          {color_on}WARNING:{color_off} Line 1
        DONE! ({color_on}1{color_off}, <Scrubbed Time>)
        """,
    ).format(
        color_on=TextwrapEx.WARNING_COLOR_ON,
        color_off=TextwrapEx.COLOR_OFF,
    )


# ----------------------------------------------------------------------
def test_WriteInfo():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteInfo("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          INFO: Line 1
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_WriteInfoWithColor():
    sink = _CreateSink(supports_colors=True)

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteInfo("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          {info_color_on}INFO:{color_off} Line 1
        DONE! ({success_color_on}0{color_off}, <Scrubbed Time>)
        """,
    ).format(
        info_color_on=TextwrapEx.INFO_COLOR_ON,
        success_color_on=TextwrapEx.SUCCESS_COLOR_ON,
        color_off=TextwrapEx.COLOR_OFF,
    )


# ----------------------------------------------------------------------
def WriteVerbose():
    # verbose == False
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteVerbose("Line 1")

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"

    # verbose == True
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(verbose=True)) as dm:
        dm.WriteVerbose("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          VERBOSE: Line 1
        DONE! (0, <Scrubbed Time>)
        """,
    )

    # verbose == True, with colors
    sink = _CreateSink(supports_colors=True)

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(verbose=True)) as dm:
        dm.WriteVerbose("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          {verbose_color_on}VERBOSE:{color_off} Line 1
        DONE! ({success_color_on}0{color_off}, <Scrubbed Time>)
        """,
    ).format(
        verbose_color_on=TextwrapEx.VERBOSE_COLOR_ON,
        success_color_on=TextwrapEx.SUCCESS_COLOR_ON,
        color_off=TextwrapEx.COLOR_OFF,
    )


# ----------------------------------------------------------------------
def test_WriteDebug():
    # debug == False
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteDebug("Line 1")

    assert _Scrub(sink.getvalue()) == "Testing...DONE! (0, <Scrubbed Time>)\n"

    # debug == True
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(debug=True)) as dm:
        dm.WriteDebug("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          DEBUG: Line 1
        DONE! (0, <Scrubbed Time>)
        """,
    )

    # debug == True, with colors
    sink = _CreateSink(supports_colors=True)

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(debug=True)) as dm:
        dm.WriteDebug("Line 1")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          {debug_color_on}DEBUG:{color_off} Line 1
        DONE! ({success_color_on}0{color_off}, <Scrubbed Time>)
        """,
    ).format(
        debug_color_on=TextwrapEx.DEBUG_COLOR_ON,
        success_color_on=TextwrapEx.SUCCESS_COLOR_ON,
        color_off=TextwrapEx.COLOR_OFF,
    )


# ----------------------------------------------------------------------
def test_AlignedMultilineOutput():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(debug=True, verbose=True)) as dm:
        content = "Line 1\nLine 2\n"

        dm.WriteLine(content)
        dm.WriteSuccess(content)
        dm.WriteError(content)
        dm.WriteWarning(content)
        dm.WriteInfo(content)
        dm.WriteVerbose(content)
        dm.WriteDebug(content)

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
          Line 2
          SUCCESS: Line 1
                   Line 2
          ERROR: Line 1
                 Line 2
          WARNING: Line 1
                   Line 2
          INFO: Line 1
                Line 2
          VERBOSE: Line 1
                   Line 2
          DEBUG: Line 1
                 Line 2
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_ErrorPrecedence():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteSuccess("Line 1")
        dm.WriteError("Line 2")
        dm.WriteWarning("Line 3")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          SUCCESS: Line 1
          ERROR: Line 2
          WARNING: Line 3
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_WarningPrecedence():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteSuccess("Line 1")
        dm.WriteWarning("Line 2")
        dm.WriteSuccess("Line 3")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          SUCCESS: Line 1
          WARNING: Line 2
          SUCCESS: Line 3
        DONE! (1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
class TestWriteStatus:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing") as dm:
            dm.WriteStatus("Line 1")

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[1A\r",
            "  ",
            "\x1b[1B",
            "\r",
            "DONE! (0, <Scrubbed Time>)\n",
        ]

    # ----------------------------------------------------------------------
    def test_NoPreserveStatusFlag(self):
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing", preserve_status=False) as dm:
            dm.WriteStatus("Line 1")

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",  # 0 - 2
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 3 - 5
            "  ",
            "\x1b[1A\r",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 7 - 9
            "  ",
            "\x1b[1A\r",
            "  ",
            "\x1b[1A\rTesting...",
            None,
            "DONE! (0, <Scrubbed Time>)\n",  # 10 - 14
        ]

    # ----------------------------------------------------------------------
    def test_Truncated(self):
        num_cols = 20

        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing", num_cols=num_cols) as dm:
            dm.WriteStatus("This is a really long status line that should be truncated.")

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "\r  This ...uncated.",
            "\n",
            "  ",
            "\x1b[1A\r",
            "  ",
            "\x1b[1B",
            "\r",
            "DONE! (0, <Scrubbed Time>)\n",
        ]

    # ----------------------------------------------------------------------
    def test_NotInteractive(self):
        stream = _FakeStream()

        with DoneManager.Create(stream, "Testing") as dm:
            dm.WriteStatus("Line 1\nLine 2")
            dm.WriteStatus("Line 3")

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "DONE! (0, <Scrubbed Time>)\n",
        ]

    # ----------------------------------------------------------------------
    def test_Nested(self):
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing") as dm:
            dm.WriteStatus("Line 1")

            with dm.Nested("Nested") as nested_dm:
                nested_dm.WriteStatus("Line 2\nLine 3")
                nested_dm.WriteStatus("Line 4")

            dm.WriteStatus("Line 5")

        stream.content[48] = _Scrub(cast(str, stream.content[48]))
        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",  # 0 - 2
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 3 - 5
            "  ",
            "\x1b[1A\r",
            "  ",
            "\x1b[1B",
            "\r",
            "  ",
            "Nested...",
            None,
            "\n",  # 6 - 14
            "  ",
            "  ",
            "\r    Line 2".ljust(Capabilities.DEFAULT_COLUMNS - len("    ") + 1),
            "\n",  # 15 - 18
            "  ",
            "  ",
            "\r    Line 3".ljust(Capabilities.DEFAULT_COLUMNS - len("    ") + 1),
            "\n",  # 19 - 22
            "  ",
            "  ",
            "\x1b[2A\r",
            "  ",
            "  ",
            "\r    Line 4".ljust(Capabilities.DEFAULT_COLUMNS - len("    ") + 1),
            "\n",  # 23 - 29
            "  ",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("    ") + 1),
            "\n",  # 30 - 33
            "  ",
            "  ",
            "\x1b[2A\r",
            "  ",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("    ") + 1),
            "\n",  # 34 - 40
            "  ",
            "  ",
            "\x1b[1A\r",
            "  ",
            "  ",
            "\x1b[1A\r  Nested...",
            None,
            "DONE! (0, <Scrubbed Time>)",
            "\n",  # 41 - 49
            None,
            "  ",
            "\r  Line 5".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 50 - 53
            "  ",
            "\x1b[1A\r",
            "  ",
            "\x1b[1B",
            "\r",
            "DONE! (0, <Scrubbed Time>)\n",  # 54 - 59
        ]

    # ----------------------------------------------------------------------
    def test_ClearStatus(self):
        # preserve_status == True
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing", preserve_status=True) as dm:
            dm.WriteStatus("Line 1\nLine 2")
            dm.ClearStatus()

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",  # 0 - 2
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 3 - 5
            "  ",
            "\r  Line 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 6 - 8
            "  ",
            "\x1b[2A\r",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 9 - 13
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",  # 14 - 16
            "  ",
            "\x1b[2A\r",
            "DONE! (0, <Scrubbed Time>)\n",  # 17 - 19
        ]

        # preserve_status == False
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing", preserve_status=False) as dm:
            dm.WriteStatus("Line 1\nLine 2")
            dm.ClearStatus()

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Line 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\x1b[1A\rTesting...",
            None,
            "DONE! (0, <Scrubbed Time>)\n",
        ]

    # ----------------------------------------------------------------------
    def test_PreserveStatus(self):
        # preserve_status == True
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing", preserve_status=True) as dm:
            dm.WriteStatus("Line 1\nLine 2")
            dm.PreserveStatus()

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Line 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\x1b[2B",
            "\r",
            "DONE! (0, <Scrubbed Time>)\n",
        ]

        # preserve_status == False
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing", preserve_status=False) as dm:
            dm.WriteStatus("Line 1\nLine 2")
            dm.PreserveStatus()

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "\r  Line 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Line 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\x1b[2B",
            "\r",
            "  ",
            "\x1b[1A\rTesting...",
            None,
            "DONE! (0, <Scrubbed Time>)\n",
        ]

    # ----------------------------------------------------------------------
    def test_ShortAndLongLines(self):
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing") as dm:
            dm.WriteStatus("Short 1\nShort 2")
            dm.WriteStatus("This is longer status 1\nThis is longer status 2\nThis is longer status 3")
            dm.WriteStatus("Again 1")

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "\r  Short 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Short 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\r  This is longer status 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  This is longer status 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  This is longer status 3".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[3A\r",
            "  ",
            "\r  Again 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[3A\r",
            "  ",
            "\x1b[1B",
            "\r",
            "DONE! (0, <Scrubbed Time>)\n",
        ]

    # ----------------------------------------------------------------------
    def test_GroupedStatus(self):
        stream = _FakeStream(
            is_interactive=True,
        )

        with DoneManager.Create(stream, "Testing") as dm:
            dm.WriteLine("One")
            dm.WriteStatus("Status 1\nStatus 2")

            dm.WriteLine("Two")
            dm.WriteStatus("Status 3\nStatus 4")

            dm.ClearStatus()

        stream.content[-1] = _Scrub(cast(str, stream.content[-1]))

        assert stream.content == [
            "Testing...",
            None,
            "\n",
            "  ",
            "One",
            "\n",
            "  ",
            "\r  Status 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Status 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "Two",
            "\n",
            "  ",
            "\r  Status 1".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Status 2".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\r  Status 3".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r  Status 4".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\r".ljust(Capabilities.DEFAULT_COLUMNS - len("  ") + 1),
            "\n",
            "  ",
            "\x1b[2A\r",
            "DONE! (0, <Scrubbed Time>)\n",
        ]


# ----------------------------------------------------------------------
def test_SimpleNested():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing 1") as dm:
        with dm.Nested("Testing 2"):
            pass

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing 1...
          Testing 2...DONE! (0, <Scrubbed Time>)
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_ComplicatedNested():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing 1") as dm:
        dm.WriteLine("1.A")

        with dm.Nested("Testing 2") as this_dm:
            this_dm.WriteLine("2.A")

        dm.WriteLine("1.B")

        with dm.Nested("Testing 3") as this_dm:
            with this_dm.Nested("3.A") as nested_dm:
                nested_dm.WriteLine("3.A.A")

        dm.WriteLine("1.C")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing 1...
          1.A
          Testing 2...
            2.A
          DONE! (0, <Scrubbed Time>)
          1.B
          Testing 3...
            3.A...
              3.A.A
            DONE! (0, <Scrubbed Time>)
          DONE! (0, <Scrubbed Time>)
          1.C
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_NestedErrors():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        assert dm.result == 0

        with dm.Nested("Zero") as this_dm:
            this_dm.result = 1

        assert dm.result == 1

        with dm.Nested("One") as this_dm:
            this_dm.result = -1

        assert dm.result == -1

        with dm.Nested("Two") as this_dm:
            this_dm.result = -2

        with dm.Nested("Three") as this_dm:
            this_dm.result = 2

        assert dm.result == -1

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Zero...DONE! (1, <Scrubbed Time>)
          One...DONE! (-1, <Scrubbed Time>)
          Two...DONE! (-2, <Scrubbed Time>)
          Three...DONE! (2, <Scrubbed Time>)
        DONE! (-1, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_VerboseNested():
    # verbose == False
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteLine("Line 1")

        with dm.VerboseNested("Verbose") as verbose_dm:
            verbose_dm.WriteLine("Line 2")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
        DONE! (0, <Scrubbed Time>)
        """,
    )

    # verbose == True
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(verbose=True)) as dm:
        dm.WriteLine("Line 1")

        with dm.VerboseNested("Verbose") as verbose_dm:
            verbose_dm.WriteLine("Line 2")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
          VERBOSE: Verbose...
          VERBOSE:   Line 2
          VERBOSE: DONE! (0, <Scrubbed Time>)
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_Debug():
    # debug == False
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        dm.WriteLine("Line 1")

        with dm.DebugNested("Debug") as debug_dm:
            debug_dm.WriteLine("Line 2")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
        DONE! (0, <Scrubbed Time>)
        """,
    )

    # debug == True
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing", flags=Flags.Create(debug=True)) as dm:
        dm.WriteLine("Line 1")

        with dm.DebugNested("Debug") as debug_dm:
            debug_dm.WriteLine("Line 2")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
          DEBUG: Debug...
          DEBUG:   Line 2
          DEBUG: DONE! (0, <Scrubbed Time>)
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
def test_YieldStream():
    sink = _CreateSink()

    with DoneManager.Create(sink, "Testing") as dm:
        with dm.YieldStream() as stream:
            stream.write("Line 1\nLine 2")

        dm.WriteLine("Line 3")

    assert _Scrub(sink.getvalue()) == textwrap.dedent(
        """\
        Testing...
          Line 1
          Line 2Line 3
        DONE! (0, <Scrubbed Time>)
        """,
    )


# ----------------------------------------------------------------------
class TestYieldVerboseStream:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        sink = _CreateSink()

        with DoneManager.Create(sink, "Testing") as dm:
            with dm.YieldVerboseStream() as stream:
                stream.write("Hello, world!")

        assert _Scrub(sink.getvalue()) == textwrap.dedent(
            """\
            Testing...DONE! (0, <Scrubbed Time>)
            """,
        )

    # ----------------------------------------------------------------------
    def test_WithVerbose(self):
        sink = _CreateSink()

        with DoneManager.Create(sink, "Testing", flags=Flags.Create(verbose=True)) as dm:
            with dm.YieldVerboseStream() as stream:
                stream.write("Hello, world!")

        assert _Scrub(sink.getvalue()) == textwrap.dedent(
            """\
            Testing...
              VERBOSE: Hello, world!DONE! (0, <Scrubbed Time>)
            """,
        )


# ----------------------------------------------------------------------
class TestYieldDebugStream:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        sink = _CreateSink()

        with DoneManager.Create(sink, "Testing") as dm:
            with dm.YieldDebugStream() as stream:
                stream.write("Hello, world!")

        assert _Scrub(sink.getvalue()) == textwrap.dedent(
            """\
            Testing...DONE! (0, <Scrubbed Time>)
            """,
        )

    # ----------------------------------------------------------------------
    def test_WithDebug(self):
        sink = _CreateSink()

        with DoneManager.Create(sink, "Testing", flags=Flags.Create(debug=True)) as dm:
            with dm.YieldDebugStream() as stream:
                stream.write("Hello, world!\n")

        assert _Scrub(sink.getvalue()) == textwrap.dedent(
            """\
            Testing...
              DEBUG: Hello, world!
            DONE! (0, <Scrubbed Time>)
            """,
        )


# ----------------------------------------------------------------------
class TestYieldStdout:
    # ----------------------------------------------------------------------
    def test_Standard(self):
        with DoneManager.Create(sys.stdout, "Testing") as dm:
            with dm.YieldStdout() as context:
                assert context.stream is sys.stdout
                assert context.line_prefix == "  "
                assert context.persist_content is True

                context.stream.write("Hello, world!\n")

                # Verify that the text appears in sys.stdout

    # ----------------------------------------------------------------------
    def test_NoPreserve(self):
        with DoneManager.Create(sys.stdout, "Testing") as dm:
            with dm.YieldStdout() as context:
                context.stream.write("Hello, world!\n")

                context.persist_content = False

                # Verify that the text does not appear in sys.stdout


# ----------------------------------------------------------------------
# |
# |  Private Types
# |
# ----------------------------------------------------------------------
class _FakeStream(TextWriter):
    # ----------------------------------------------------------------------
    def __init__(
        self,
        *,
        fileno: int = 123,
        isatty: bool = False,
        is_headless: bool = True,
        is_interactive: bool = False,
        supports_colors: bool = False,
    ):
        self.content: list[Optional[str]] = []
        self._fileno = fileno
        self._isatty = isatty

        Capabilities(
            stream=self,
            is_headless=is_headless,
            is_interactive=is_interactive,
            supports_colors=supports_colors,
        )

    # ----------------------------------------------------------------------
    def isatty(self) -> bool:
        return self._isatty

    # ----------------------------------------------------------------------
    def write(
        self,
        content: str,
    ) -> int:
        self.content.append(content)
        return len(content)

    # ----------------------------------------------------------------------
    def flush(self) -> None:
        self.content.append(None)

    # ----------------------------------------------------------------------
    def fileno(self) -> int:
        return self._fileno

    # ----------------------------------------------------------------------
    def close(self) -> None:
        pass


# ----------------------------------------------------------------------
# |
# |  Private Functions
# |
# ----------------------------------------------------------------------
def _CreateSink(
    supports_colors: bool = False,
) -> StringIO:
    sink = StringIO()

    Capabilities(
        stream=sink,
        is_headless=True,
        is_interactive=False,
        supports_colors=supports_colors,
    )

    return sink


# ----------------------------------------------------------------------
def _Scrub(
    content: str,
) -> str:
    return re.sub(
        r"\d+:\d{2}:\d{2}(?:\.\d+)?",
        "<Scrubbed Time>",
        content,
    )



# ----------------------------------------------------------------------
# |
# |  TextwrapEx.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-19 15:40:12
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Enhancements for the textwrap library."""

import math
import textwrap

from enum import auto, Enum
from typing import Callable, Optional


# ----------------------------------------------------------------------
# |
# |  Public Types
# |
# ----------------------------------------------------------------------
# Normally, I'd use colorama for these values, but this package cannot have any
# external dependencies.
BRIGHT_RED_COLOR_ON = "\033[31;1m"  # Red / Bright
BRIGHT_GREEN_COLOR_ON = "\033[32;1m"  # Green / Bright
BRIGHT_YELLOW_COLOR_ON = "\033[33;1m"  # Yellow / Bright

BRIGHT_WHITE_COLOR_ON = "\033[37;1m"  # White / Bright
DIM_WHITE_COLOR_ON = "\033[;7m"  # Inverse video

# ----------------------------------------------------------------------
ERROR_COLOR_ON = BRIGHT_RED_COLOR_ON
INFO_COLOR_ON = DIM_WHITE_COLOR_ON
SUCCESS_COLOR_ON = BRIGHT_GREEN_COLOR_ON
WARNING_COLOR_ON = BRIGHT_YELLOW_COLOR_ON
VERBOSE_COLOR_ON = DIM_WHITE_COLOR_ON
DEBUG_COLOR_ON = BRIGHT_WHITE_COLOR_ON

# ----------------------------------------------------------------------
COLOR_OFF = "\033[0m"  # Reset


# ----------------------------------------------------------------------
class Justify(Enum):
    """Line justification"""

    # ----------------------------------------------------------------------
    Left = auto()
    Center = auto()
    Right = auto()

    # ----------------------------------------------------------------------
    def Justify(
        self,
        value: str,
        padding: int,
    ) -> str:
        if self == Justify.Left:
            return value.ljust(padding)

        if self == Justify.Center:
            return value.center(padding)

        if self == Justify.Right:
            return value.rjust(padding)

        assert False, self  # pragma: no cover


# ----------------------------------------------------------------------
# |
# |  Public Functions
# |
# ----------------------------------------------------------------------
def _CreateCustomPrefixFunc(
    header: str,
    color_value: str,
) -> Callable[
    [
        bool,
    ],
    str,
]:  # supports_colors
    # ----------------------------------------------------------------------
    def Impl(
        supports_colors: bool,
    ) -> str:
        if supports_colors:
            this_color_on = color_value
            this_color_off = COLOR_OFF
        else:
            this_color_on = ""
            this_color_off = ""

        return "{}{}:{} ".format(this_color_on, header, this_color_off)

    # ----------------------------------------------------------------------

    return Impl


# ----------------------------------------------------------------------
CreateErrorPrefix = _CreateCustomPrefixFunc("ERROR", ERROR_COLOR_ON)
CreateWarningPrefix = _CreateCustomPrefixFunc("WARNING", WARNING_COLOR_ON)
CreateInfoPrefix = _CreateCustomPrefixFunc("INFO", INFO_COLOR_ON)
CreateSuccessPrefix = _CreateCustomPrefixFunc("SUCCESS", SUCCESS_COLOR_ON)
CreateVerbosePrefix = _CreateCustomPrefixFunc("VERBOSE", VERBOSE_COLOR_ON)
CreateDebugPrefix = _CreateCustomPrefixFunc("DEBUG", DEBUG_COLOR_ON)

del _CreateCustomPrefixFunc


# ----------------------------------------------------------------------
def CreateErrorText(
    value: str,
    *,
    supports_colors: bool = True,
    decorate_every_line: bool = False,
) -> str:
    return _CreateText(
        CreateErrorPrefix,
        value,
        supports_colors=supports_colors,
        decorate_every_line=decorate_every_line,
    )


# ----------------------------------------------------------------------
def CreateWarningText(
    value: str,
    *,
    supports_colors: bool = True,
    decorate_every_line: bool = False,
) -> str:
    return _CreateText(
        CreateWarningPrefix,
        value,
        supports_colors=supports_colors,
        decorate_every_line=decorate_every_line,
    )


# ----------------------------------------------------------------------
def CreateInfoText(
    value: str,
    *,
    supports_colors: bool = True,
    decorate_every_line: bool = False,
) -> str:
    return _CreateText(
        CreateInfoPrefix,
        value,
        supports_colors=supports_colors,
        decorate_every_line=decorate_every_line,
    )


# ----------------------------------------------------------------------
def CreateSuccessText(
    value: str,
    *,
    supports_colors: bool = True,
    decorate_every_line: bool = False,
) -> str:
    return _CreateText(
        CreateSuccessPrefix,
        value,
        supports_colors=supports_colors,
        decorate_every_line=decorate_every_line,
    )


# ----------------------------------------------------------------------
def CreateVerboseText(
    value: str,
    *,
    supports_colors: bool = True,
    decorate_every_line: bool = False,
) -> str:
    return _CreateText(
        CreateVerbosePrefix,
        value,
        supports_colors=supports_colors,
        decorate_every_line=decorate_every_line,
    )


# ----------------------------------------------------------------------
def CreateDebugText(
    value: str,
    *,
    supports_colors: bool = True,
    decorate_every_line: bool = False,
) -> str:
    return _CreateText(
        CreateDebugPrefix,
        value,
        supports_colors=supports_colors,
        decorate_every_line=decorate_every_line,
    )


# ----------------------------------------------------------------------
def CreateStatusText(
    succeeded: Optional[int],
    failed: Optional[int],
    warnings: Optional[int],
    *,
    supports_colors: bool = True,
) -> str:
    """Returns text that contains information for the number of succeeded, failed, and warning items which is useful when displaying status information."""

    if supports_colors:
        success_on = SUCCESS_COLOR_ON
        failed_on = ERROR_COLOR_ON
        warning_on = WARNING_COLOR_ON
        color_off = COLOR_OFF
    else:
        success_on = ""
        failed_on = ""
        warning_on = ""
        color_off = ""

    parts: list[str] = []

    if succeeded is not None:
        if succeeded == 0:
            prefix = "0"
        else:
            prefix = "{}{}{}".format(success_on, succeeded, color_off)

        parts.append("{} succeeded".format(prefix))

    if failed is not None:
        if failed == 0:
            prefix = "0"
        else:
            prefix = "{}{}{}".format(failed_on, failed, color_off)

        parts.append("{} failed".format(prefix))

    if warnings is not None:
        if warnings == 0:
            prefix = "0"
        else:
            prefix = "{}{}{}".format(warning_on, warnings, color_off)

        parts.append("{} warnings".format(prefix))

    value = ", ".join(parts)

    return value


# ----------------------------------------------------------------------
def Indent(
    value: str,
    indentation: str | int,
    *,
    skip_first_line: bool = False,
) -> str:
    """Ensures that each line in the provided value contains  the specified indentation."""

    if isinstance(indentation, int):
        indentation = " " * indentation
    elif isinstance(indentation, str):
        # Nothing to do here
        pass
    else:
        assert False, indentation  # pragma: no cover

    if skip_first_line:
        is_first_line = True

        # ----------------------------------------------------------------------
        def ShouldIndent(_) -> bool:
            nonlocal is_first_line

            if is_first_line:
                is_first_line = False
                return False

            return True

        # ----------------------------------------------------------------------

        should_indent = ShouldIndent
    else:
        should_indent = lambda _: True

    assert isinstance(indentation, str), indentation
    return textwrap.indent(
        value,
        indentation,
        lambda content: not content.isspace() and should_indent(content),
    )


# ----------------------------------------------------------------------
def BoundedLJust(
    value: str,
    length: int,
) -> str:
    """\
    Returns a string that is length chars long. Content will be left-justified or
    trimmed (with an ellipsis) if necessary.
    """

    len_value = len(value)

    if len_value < length:
        value = value.ljust(length)
    elif len_value > length:
        chars_to_trim = (len_value - length + 3) / 2
        midpoint = math.floor(len_value / 2)

        # Ensure a consistent ellipsis placement
        if not length & 1 and not len_value & 1:
            midpoint -= 1

        value = "{}...{}".format(
            value[: (midpoint - math.floor(chars_to_trim))],
            value[(midpoint + math.ceil(chars_to_trim)) :],
        )

    return value


# ----------------------------------------------------------------------
def CreateAnsiHyperLink(
    url: str,
    value: str,
) -> str:
    return "\033]8;;{}\033\\{}\033]8;;\033\\".format(url, value)


# ----------------------------------------------------------------------
def CreateTable(
    headers: list[str],
    all_values: list[list[str]],
    col_justifications: list[Justify] | None = None,
    decorate_values_func: Callable[[int, list[str]], list[str]] | None = None,
    on_col_sizes_calculated: Callable[[list[int]], None] | None = None,
    col_padding: str = "  ",
    *,
    decorate_headers: bool = False,
) -> str:
    """Prints a table with the provided headers and values."""

    assert col_justifications is None or len(col_justifications) == len(headers)
    assert decorate_headers is False or decorate_values_func

    col_justifications = col_justifications or [Justify.Left] * len(headers)
    decorate_values_func = decorate_values_func or (lambda _, row: row)
    on_col_sizes_calculated = on_col_sizes_calculated or (lambda _: None)

    # Calculate the col sizes
    col_sizes = [len(header) for header in headers]

    # Get the column size for each row
    for row in all_values:
        assert len(row) == len(headers)
        for index, col_value in enumerate(row):
            col_sizes[index] = max(len(col_value), col_sizes[index])

    on_col_sizes_calculated(col_sizes)

    # Create the template
    row_template = col_padding.join(
        "{{:<{}}}".format(col_size) if col_size != 0 else "{}" for col_size in col_sizes
    )

    # Create the rows
    rows: list[str] = []

    # ----------------------------------------------------------------------
    def CreateRow(
        index: int,
        values: list[str],
    ) -> None:
        decorated_values: list[str] = []

        for col_justification, col_value, col_size in zip(col_justifications, values, col_sizes):
            decorated_values.append(col_justification.Justify(col_value, col_size))

        if index >= 0 or decorate_headers:
            decorated_values = decorate_values_func(index, decorated_values)

        rows.append(row_template.format(*decorated_values).rstrip())

    # ----------------------------------------------------------------------

    CreateRow(-2, headers)
    CreateRow(-1, ["-" * col_size for col_size in col_sizes])

    for index, values in enumerate(all_values):
        CreateRow(index, values)

    return "\n".join(rows) + "\n"


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
def _CreateText(
    create_prefix_func: Callable[
        [
            bool,  # supports_colors
        ],
        str,
    ],
    value: str,
    *,
    supports_colors: bool,
    decorate_every_line: bool,
) -> str:
    prefix = create_prefix_func(supports_colors)

    if decorate_every_line:
        return Indent(value, prefix)

    # Put newlines before the header
    starting_index = 0

    while starting_index < len(value) and value[starting_index] == "\n":
        starting_index += 1

    return "{}{}".format(
        value[:starting_index],
        Indent(
            "{}{}".format(prefix, value[starting_index:]),
            # The indent is the length of the prefix without color decorations
            len(create_prefix_func(False)),
            skip_first_line=True,
        ),
    )



# ----------------------------------------------------------------------
# |
# |  StreamDecorator.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2023-12-18 18:11:38
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2023-24
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Contains the StreamDecorator object"""

import sys

from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import auto, Enum
from functools import cached_property
from typing import Callable, Generator, Iterator, Optional

from dbrownell_Common.ContextlibEx import ExitStack
from dbrownell_Common.Streams.Capabilities import Capabilities, TextWriterT
from dbrownell_Common.Streams.TextWriter import TextWriter


# ----------------------------------------------------------------------
class StreamDecorator(TextWriter):
    """Stream-like object that supports line decoration with prefixes and suffixes."""

    # ----------------------------------------------------------------------
    # |
    # |  Public Types
    # |
    # ----------------------------------------------------------------------
    PrefixOrSuffixT = (
        None
        | str
        | Callable[
            [
                int,  # column offset
            ],
            str,
        ]
    )

    # ----------------------------------------------------------------------
    # |
    # |  Public Methods
    # |
    # ----------------------------------------------------------------------
    def __init__(
        self,
        stream_or_streams: None | TextWriterT | list[TextWriterT],
        line_prefix: "StreamDecorator.PrefixOrSuffixT" = None,
        line_suffix: "StreamDecorator.PrefixOrSuffixT" = None,
        prefix: "StreamDecorator.PrefixOrSuffixT" = None,
        suffix: "StreamDecorator.PrefixOrSuffixT" = None,
        *,
        decorate_empty_lines: bool = True,
    ):
        if stream_or_streams is None:
            streams = []
        elif isinstance(stream_or_streams, list):
            streams = stream_or_streams
        else:
            streams = [
                stream_or_streams,
            ]

        del stream_or_streams

        # ----------------------------------------------------------------------
        def PrefixOrSuffixTToCallable(
            value: StreamDecorator.PrefixOrSuffixT,
        ) -> Callable[[int], str]:
            if value is None:
                return lambda _: ""

            if isinstance(value, str):
                return lambda _: value

            if callable(value):
                return value

            assert False, value  # pragma: no cover

        # ----------------------------------------------------------------------

        self._streams = streams
        self._line_prefix_func = PrefixOrSuffixTToCallable(line_prefix)
        self._line_suffix_func = PrefixOrSuffixTToCallable(line_suffix)
        self._prefix_func = PrefixOrSuffixTToCallable(prefix)
        self._suffix_func = PrefixOrSuffixTToCallable(suffix)
        self._decorate_empty_lines = decorate_empty_lines

        self._state: StreamDecorator._State = StreamDecorator._State.Prefix
        self._col_offset = 0
        self._wrote_content = False

        self._isatty = any(getattr(stream, "isatty", lambda: False)() for stream in self._streams)

        # Set the capabilities for this stream
        if self._streams:
            lowest_capabilities: Optional[Capabilities] = None

            for stream in self._streams:
                capabilities = Capabilities.Get(stream)

                if lowest_capabilities is None or capabilities < lowest_capabilities:
                    lowest_capabilities = capabilities

            assert lowest_capabilities is not None

            capabilities = lowest_capabilities

        else:
            capabilities = Capabilities(ignore_environment=True)

        Capabilities.Set(self, capabilities)

    # ----------------------------------------------------------------------
    @property
    def col_offset(self) -> int:
        return self._col_offset

    @cached_property
    def has_stdout(self) -> bool:
        return any(
            stream is sys.stdout or (isinstance(stream, StreamDecorator) and stream.has_stdout)
            for stream in self._streams
        )

    @cached_property
    def has_streams(self) -> bool:
        return bool(self._streams)

    @property
    def wrote_content(self) -> bool:
        return self._wrote_content

    # ----------------------------------------------------------------------
    def isatty(self) -> bool:
        return self._isatty

    # ----------------------------------------------------------------------
    def fileno(self) -> int:
        return next((stream.fileno() for stream in self._streams if hasattr(stream, "fileno")), 0)  # type: ignore

    # ----------------------------------------------------------------------
    def write(
        self,
        content: str,
    ) -> int:
        if self._state == StreamDecorator._State.Closed:
            raise Exception("Instance is closed.")

        chars_written = 0

        if self._state == StreamDecorator._State.Prefix:
            self._write_raw(self._prefix_func(self._col_offset))
            self._state = StreamDecorator._State.Writing

        chars_written += self._write_content(content)

        return chars_written

    # ----------------------------------------------------------------------
    def flush(self) -> None:
        if self._state == StreamDecorator._State.Closed:
            raise Exception("Instance is closed.")

        for stream in self._streams:
            stream.flush()

    # ----------------------------------------------------------------------
    def close(self) -> None:
        if self._state == StreamDecorator._State.Closed:
            raise Exception("Instance is closed.")

        self._state = StreamDecorator._State.Suffix
        self._write_raw(self._suffix_func(self._col_offset))

        self.flush()

        self._state = StreamDecorator._State.Closed

        for stream in self._streams:
            stream.close()

    # ----------------------------------------------------------------------
    def EnumStreams(self) -> Generator[TextWriterT, None, None]:
        yield from self._streams

    # ----------------------------------------------------------------------
    def GetLinePrefix(
        self,
        column: int,
    ) -> str:
        return self._line_prefix_func(column)

    # ----------------------------------------------------------------------
    def GetCompleteLinePrefix(
        self,
        *,
        include_self: bool = True,
    ) -> str:
        prefixes = self._GetLinePrefixInfo(0, include_self=include_self)[1]

        return "".join(prefixes)

    # ----------------------------------------------------------------------
    @dataclass
    class YieldStdoutContext:
        stream: TextWriterT
        line_prefix: str
        persist_content: bool = field(kw_only=True)

    @contextmanager
    def YieldStdout(self) -> Iterator[YieldStdoutContext]:
        """Provides access to a stdout stream in a way that doesn't impact the indentation level maintained by a hierarchy of StreamDecorators."""

        line_prefix = self.GetCompleteLinePrefix()

        context = StreamDecorator.YieldStdoutContext(sys.stdout, line_prefix, persist_content=True)

        if not Capabilities.Get(self).is_interactive:
            with ExitStack(lambda: self.write("\n")):
                self.write("\n")
                yield context

            return

        if not self.has_stdout:
            raise Exception(
                "This functionality can only be used with streams that ultimately write to `sys.stdout`"
            )

        yield context

    # ----------------------------------------------------------------------
    # |
    # |  Private Types
    # |
    # ----------------------------------------------------------------------
    class _State(Enum):
        Prefix = auto()
        Writing = auto()
        Suffix = auto()
        Closed = auto()

    # ----------------------------------------------------------------------
    # |
    # |  Private Methods
    # |
    # ----------------------------------------------------------------------
    def _write_content(
        self,
        content: str,
    ) -> int:
        chars_written = 0

        if not content:
            return chars_written

        for line, has_newline in self.__class__._EnumLines(  # pylint: disable=protected-access
            content
        ):  # pylint: disable=protected-access
            if self._col_offset == 0 and (line or self._decorate_empty_lines):
                chars_written += self._write_raw(self._line_prefix_func(self._col_offset))

            chars_written += self._write_raw(line)

            if has_newline:
                if self._col_offset != 0 or self._decorate_empty_lines:
                    chars_written += self._write_raw(self._line_suffix_func(self._col_offset))

                chars_written += self._write_raw("\n")

        return chars_written

    # ----------------------------------------------------------------------
    def _write_raw(
        self,
        content: str,
    ) -> int:
        if not content:
            return 0

        for stream in self._streams:
            try:
                stream.write(content)

            except UnicodeEncodeError as ex:
                wrote_content = False

                for decode_error_method in [
                    "surrogateescape",
                    "replace",
                    "backslashreplace",
                    "ignore",
                ]:
                    try:
                        decoded_content = content.encode("utf-8").decode("ascii", decode_error_method)
                        stream.write(decoded_content)

                        wrote_content = True
                        break
                    except UnicodeEncodeError:
                        pass

                if not wrote_content:
                    raise ex

        self._wrote_content = True

        # Find the last '\r' or '\n'
        len_content = len(content)
        index = len_content - 1

        while index >= 0:
            if content[index] in ("\r", "\n"):
                break

            index -= 1

        index += 1

        if index == 0:
            self._col_offset += len(content)
        else:
            self._col_offset = len(content[index:])

        return len(content)

    # ----------------------------------------------------------------------
    @staticmethod
    def _EnumLines(
        content: str,
    ) -> Generator[tuple[str, bool], None, None]:
        # Scenarios:
        #   '\n' => ['']
        #   'a' => ['a']
        #   'a\nb' => ['a', 'b']
        #   'a\nb\n' => ['a', 'b', '']

        lines: list[str] = content.split("\n")

        if len(lines) == 1:
            has_newline = not lines[0]

            yield lines[0], has_newline
            return

        if not lines[-1]:
            final_newline = True
            del lines[-1]
        else:
            final_newline = False

        last_index = len(lines) - 1

        for index, line in enumerate(lines):
            has_newline = index < last_index or final_newline

            yield line, has_newline

    # ----------------------------------------------------------------------
    def _GetLinePrefixInfo(
        self,
        column: int,
        *,
        include_self: bool,
    ) -> tuple[
        int,
        list[str],
    ]:  # column offset  # prefix values
        if not self._streams:
            return column, []

        if isinstance(self._streams[0], StreamDecorator):
            column, prefixes = self._streams[  # pylint: disable=protected-access
                0
            ]._GetLinePrefixInfo(  # pylint: disable=protected-access
                column, include_self=True
            )
        else:
            prefixes = []

        if include_self:
            prefix = self.GetLinePrefix(column)
            if prefix:
                prefixes.append(prefix)
                column += len(prefix)

        return column, prefixes



# ----------------------------------------------------------------------
# |
# |  ExecuteTasks_TestManual.py
# |
# |  David Brownell <db@DavidBrownell.com>
# |      2024-01-13 20:50:28
# |
# ----------------------------------------------------------------------
# |
# |  Copyright David Brownell 2024
# |  Distributed under the MIT License.
# |
# ----------------------------------------------------------------------
"""Manual tests for ExecuteTasks.py."""

# Testing Challenges
# ==================
# `ExecuteTasks.py` has proven to be very difficult to test for a couple of reasons:
#
# 1) The implementation executes across a number of threads which means the results are non-
#    deterministic.
#
# 2) The implementation relies on sys.stdout, the interactive nature of streams, and whether
#    a stream is headless or not. The way in which a test is run impacts all of these.
#
# Because of these factors, I have broken the tests into different categories:
#
# A) `ExecuteTasks_UnitTest.py`: Those things that can be verified automatically.
# B) `ExecuteTasks_TestManual.py`: Those things that cannot be verified automatically.
#
# Both of these tests should be run when introducing changes in `ExecuteTasks.py`.
#
# I wish that I knew of a better way to write these tests so that they didn't have to be run across
# multiple files and were fully automated. However, I have not been able to figure out a way to do
# that yet.
#
# To Run the Tests
# ================
#
# `ExecuteTasks_UnitTest.py`
# --------------------------
#
# Standard:
# `python Build.py Test`
#
# Code Coverage:
# `python Build.py Test --code-coverage`
#
# `ExecuteTasks_TestManual.py`
# ----------------------------
#
# Standard:
# `pytest tests/dbrownell_Common/ExecuteTasks_TestManual.py -vv --capture=no`
#
# Code Coverage:
# `pytest tests/dbrownell_Common/ExecuteTasks_TestManual.py -vv --capture=no --cov=dbrownell_Common.ExecuteTasks`
#
# Code Coverage with lcov output:
# `pytest tests/dbrownell_Common/ExecuteTasks_TestManual.py -vv --capture=no --cov=dbrownell_Common.ExecuteTasks --cov-report=lcov:tests\dbrownell_Common\lcov.info`
#

import multiprocessing
import sys
import time

from io import StringIO
from pathlib import Path
from typing import Any, Callable, cast

import pytest

from dbrownell_Common.ExecuteTasks import *
from dbrownell_Common import PathEx
from dbrownell_Common.Streams.DoneManager import DoneManager, Flags


# ----------------------------------------------------------------------
@pytest.mark.parametrize("is_interactive", [True, False])
def test_Standard(is_interactive):
    sys.stdout.write("\n")

    if is_interactive:
        stream = sys.stdout
    else:
        stream = StringIO()

    num_tasks = int(multiprocessing.cpu_count() * 1.5)
    num_steps = 20

    with DoneManager.Create(
        stream,
        "Executing tasks",
        flags=Flags.Create(verbose=True),
    ) as dm:
        # ----------------------------------------------------------------------
        def Execute(context, status):
            for x in range(num_steps):
                status.OnProgress(x, "Iteration {}".format(x))
                status.SetTitle("{}.{}".format(context, x))

                time.sleep(0.25)

            if context % 5 == 0:
                status.OnInfo("{} is a multiple of 5".format(context))
                return context * 2, 0

            if context & 1:
                status.OnInfo("{} is odd".format(context), verbose=True)
                return context * 2, 1

            return context * 2, -1

        # ----------------------------------------------------------------------

        results = _ExecuteTasks(
            dm,
            Execute,
            num_tasks=num_tasks,
            num_steps=num_steps,
        )

    for index, result in enumerate(results):
        assert result == index * 2

    sys.stdout.write("\n")


# ----------------------------------------------------------------------
def test_SingleThread():
    sys.stdout.write("\n")

    num_tasks = 5
    num_steps = 10

    with DoneManager.Create(sys.stdout, "Executing tasks...") as dm:
        # ----------------------------------------------------------------------
        def Execute(context, status):
            for x in range(num_steps):
                status.OnProgress(x, "Iteration {}".format(x))
                time.sleep(0.25)

            return context * 2, 0

        # ----------------------------------------------------------------------

        results = _ExecuteTasks(
            dm,
            Execute,
            num_tasks=num_tasks,
            num_steps=num_steps,
            max_num_threads=1,
        )

    for index, result in enumerate(results):
        assert result == index * 2

    sys.stdout.write("\n")


# ----------------------------------------------------------------------
def test_YieldQueueExecutor():
    sys.stdout.write("\n")

    num_tasks = int(multiprocessing.cpu_count() * 1.5)
    num_steps = 20

    results: list[Optional[int]] = [
        None,
    ] * num_tasks

    with DoneManager.Create(sys.stdout, "Executing tasks...") as dm:
        with YieldQueueExecutor(dm, "Queue") as executor:
            for index in range(num_tasks):
                # ----------------------------------------------------------------------
                def Prepare(
                    on_simple_status_func: Callable[[str], None],  # pylint: disable=unused-argument
                    index=index,
                ) -> tuple[int, YieldQueueExecutorTypes.ExecuteFuncType]:
                    # ----------------------------------------------------------------------
                    def Execute(status: Status) -> Optional[str]:
                        for x in range(num_steps):
                            status.OnProgress(x, "Iteration {}".format(x))
                            time.sleep(0.25)

                        results[index] = index * 2

                    # ----------------------------------------------------------------------

                    return num_steps, Execute

                # ----------------------------------------------------------------------

                executor(str(index), Prepare)

    for index, result in enumerate(results):
        assert result == index * 2

    sys.stdout.write("\n")


# ----------------------------------------------------------------------
# |
# |  Private Functions
# |
# ----------------------------------------------------------------------
def _ExecuteTasks(
    dm: DoneManager,
    execute_func: Callable[
        [int, Status],  # context
        tuple[int, int],  # updated context, return code
    ],
    *,
    num_tasks: int = 5,
    num_steps: Optional[int] = None,
    **execute_tasks_kwargs: Any,
) -> list[int]:
    task_data = [
        TaskData(
            str(value),
            value,
            None,
        )
        for value in range(num_tasks)
    ]

    results = [
        None,
    ] * len(task_data)

    # ----------------------------------------------------------------------
    def Init(context: Any) -> tuple[Path, ExecuteTasksTypes.PrepareFuncType]:
        # ----------------------------------------------------------------------
        def Prepare(
            on_simple_status_func: Callable[[str], None],
        ) -> ExecuteTasksTypes.ExecuteFuncType | tuple[int, ExecuteTasksTypes.ExecuteFuncType]:
            # ----------------------------------------------------------------------
            def Execute(status: Status) -> int:
                result, return_code = execute_func(context, status)

                results[context] = result
                return return_code

            # ----------------------------------------------------------------------

            if num_steps is None:
                return Execute

            return num_steps, Execute

        # ----------------------------------------------------------------------

        return PathEx.CreateTempFileName(), Prepare

    # ----------------------------------------------------------------------

    ExecuteTasks(dm, "Tasks", task_data, Init, **execute_tasks_kwargs)

    return cast(list[int], results)
